# sage_setup: distribution = sagemath-modules
"""
Steenrod algebra bases

AUTHORS:

- John H. Palmieri (2008-07-30): version 0.9
- John H. Palmieri (2010-06-30): version 1.0
- Simon King (2011-10-25): Fix the use of cached functions

This package defines functions for computing various bases of the
Steenrod algebra, and for converting between the Milnor basis and
any other basis.

This packages implements a number of different bases, at least at
the prime 2. The Milnor and Serre-Cartan bases are the most
familiar and most standard ones, and all of the others are defined
in terms of one of these. The bases are described in the
documentation for the function
:func:`steenrod_algebra_basis`; also see the papers by
Monks [Mon1998]_ and Wood [Woo1998]_ for more information about them. For
commutator bases, see the preprint by Palmieri and Zhang [PZ2008]_.

- ``'milnor'`` -- Milnor basis

- ``'serre-cartan'`` or ``'adem'`` or ``'admissible'`` -- Serre-Cartan basis

Most of the rest of the bases are only defined when `p=2`.  The only
exceptions are the `P^s_t`-bases and the commutator bases, which are
defined at all primes.

- ``'wood_y'`` -- Wood's Y basis

- ``'wood_z'`` -- Wood's Z basis

- ``'wall'``, ``'wall_long'`` -- Wall's basis

- ``'arnon_a'``, ``'arnon_a_long'`` -- Arnon's A basis

- ``'arnon_c'`` -- Arnon's C basis

- ``'pst'``, ``'pst_rlex'``, ``'pst_llex'``, ``'pst_deg'``, ``'pst_revz'`` --
  various `P^s_t`-bases

- ``'comm'``, ``'comm_rlex'``, ``'comm_llex'``, ``'comm_deg'``, ``'comm_revz'``,
   or these with ``'_long'`` appended -- various commutator bases

The main functions provided here are

- :func:`steenrod_algebra_basis`.  This computes a tuple representing
  basis elements for the Steenrod algebra in a given degree, at a
  given prime, with respect to a given basis. It is a cached function.

- :func:`convert_to_milnor_matrix`.  This returns the change-of-basis
  matrix, in a given degree, from any basis to the Milnor basis. It is
  a cached function.

- :func:`convert_from_milnor_matrix`.  This returns the inverse of the
  previous matrix.

INTERNAL DOCUMENTATION:

If you want to implement a new basis for the Steenrod algebra:

In the file :file:`steenrod_algebra.py`:

For the class :class:`SteenrodAlgebra_generic
<sage.algebras.steenrod.steenrod_algebra.SteenrodAlgebra_generic>`, add functionality to the
methods:

- :meth:`_repr_term <sage.algebras.steenrod.steenrod_algebra.SteenrodAlgebra_generic._repr_term>`

- :meth:`degree_on_basis <sage.algebras.steenrod.steenrod_algebra.SteenrodAlgebra_generic.degree_on_basis>`

- :meth:`_milnor_on_basis <sage.algebras.steenrod.steenrod_algebra.SteenrodAlgebra_generic._milnor_on_basis>`

- :meth:`an_element <sage.algebras.steenrod.steenrod_algebra.SteenrodAlgebra_generic.an_element>`

In the file :file:`steenrod_algebra_misc.py`:

- add functionality to :func:`get_basis_name
  <sage.algebras.steenrod.steenrod_algebra_misc.get_basis_name>`: this
  should accept as input various synonyms for the basis, and its
  output should be a canonical name for the basis.

- add a function ``BASIS_mono_to_string`` like
  :func:`milnor_mono_to_string
  <sage.algebras.steenrod.steenrod_algebra_misc.milnor_mono_to_string>`
  or one of the other similar functions.

In this file :file:`steenrod_algebra_bases.py`:

- add appropriate lines to :func:`steenrod_algebra_basis`.

- add a function to compute the basis in a given dimension (to be
  called by :func:`steenrod_algebra_basis`).

- modify :func:`steenrod_basis_error_check` so it checks the new
  basis.

If the basis has an intrinsic way of defining a product, implement it
in the file :file:`steenrod_algebra_mult.py` and also in the
:meth:`product_on_basis
<sage.algebras.steenrod.steenrod_algebra.SteenrodAlgebra_generic.product_on_basis>`
method for :class:`SteenrodAlgebra_generic
<sage.algebras.steenrod.steenrod_algebra.SteenrodAlgebra_generic>` in
:file:`steenrod_algebra.py`.
"""

# ****************************************************************************
#  Copyright (C) 2008-2010 John H. Palmieri <palmieri@math.washington.edu>
#  Distributed under the terms of the GNU General Public License (GPL)
# ****************************************************************************

from sage.misc.cachefunc import cached_function


@cached_function
def convert_to_milnor_matrix(n, basis, p=2, generic='auto'):
    r"""
    Change-of-basis matrix, 'basis' to Milnor, in dimension
    `n`, at the prime `p`.

    INPUT:

    - ``n`` -- nonnegative integer, the dimension
    - ``basis`` -- string, the basis from which to convert
    - ``p`` -- positive prime number (default: 2)

    OUTPUT:

    ``matrix`` -- change-of-basis matrix, a square matrix over `\GF{p}`

    EXAMPLES::

        sage: # needs sage.combinat sage.modules
        sage: from sage.algebras.steenrod.steenrod_algebra_bases import convert_to_milnor_matrix
        sage: convert_to_milnor_matrix(5, 'adem')  # indirect doctest
        [0 1]
        [1 1]
        sage: convert_to_milnor_matrix(45, 'milnor')
        111 x 111 dense matrix over Finite Field of size 2 (use the '.str()' method to see the entries)
        sage: convert_to_milnor_matrix(12, 'wall')
        [1 0 0 1 0 0 0]
        [1 1 0 0 0 1 0]
        [0 1 0 1 0 0 0]
        [0 0 0 1 0 0 0]
        [1 1 0 0 1 0 0]
        [0 0 1 1 1 0 1]
        [0 0 0 0 1 0 1]

    The function takes an optional argument, the prime `p` over
    which to work::

        sage: # needs sage.combinat sage.modules
        sage: convert_to_milnor_matrix(17, 'adem', 3)
        [0 0 1 1]
        [0 0 0 1]
        [1 1 1 1]
        [0 1 0 1]
        sage: convert_to_milnor_matrix(48, 'adem', 5)
        [0 1]
        [1 1]
        sage: convert_to_milnor_matrix(36, 'adem', 3)
        [0 0 1]
        [0 1 0]
        [1 2 0]
    """
    from sage.matrix.constructor import matrix
    from sage.rings.finite_rings.finite_field_constructor import GF
    from .steenrod_algebra import SteenrodAlgebra
    if generic == 'auto':
        generic = p != 2
    if n == 0:
        return matrix(GF(p), 1, 1, [[1]])
    milnor_base = steenrod_algebra_basis(n, 'milnor', p, generic=generic)
    rows = []
    A = SteenrodAlgebra(basis=basis, p=p, generic=generic)
    for poly in A.basis(n):
        d = poly.milnor().monomial_coefficients()
        for v in milnor_base:
            entry = d.get(v, 0)
            rows = rows + [entry]
    d = len(milnor_base)
    return matrix(GF(p), d, d, rows)


def convert_from_milnor_matrix(n, basis, p=2, generic='auto'):
    r"""
    Change-of-basis matrix, Milnor to ``basis``, in dimension `n`.

    INPUT:

    - ``n`` -- nonnegative integer, the dimension

    - ``basis`` -- string, the basis to which to convert

    - ``p`` -- positive prime number (default: 2)

    OUTPUT:

    ``matrix`` -- change-of-basis matrix, a square matrix over `\GF{p}`

    .. NOTE::

        This is called internally.  It is not intended for casual
        users, so no error checking is made on the integer `n`, the
        basis name, or the prime.

    EXAMPLES::

        sage: # needs sage.combinat sage.modules
        sage: from sage.algebras.steenrod.steenrod_algebra_bases import convert_from_milnor_matrix, convert_to_milnor_matrix
        sage: convert_from_milnor_matrix(12, 'wall')
        [1 0 0 1 0 0 0]
        [0 0 1 1 0 0 0]
        [0 0 0 1 0 1 1]
        [0 0 0 1 0 0 0]
        [1 0 1 0 1 0 0]
        [1 1 1 0 0 0 0]
        [1 0 1 0 1 0 1]
        sage: convert_from_milnor_matrix(38, 'serre_cartan')
        72 x 72 dense matrix over Finite Field of size 2 (use the '.str()' method to see the entries)
        sage: x = convert_to_milnor_matrix(20, 'wood_y')
        sage: y = convert_from_milnor_matrix(20, 'wood_y')
        sage: x*y
        [1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0]
        [0 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0]
        [0 0 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0]
        [0 0 0 1 0 0 0 0 0 0 0 0 0 0 0 0 0]
        [0 0 0 0 1 0 0 0 0 0 0 0 0 0 0 0 0]
        [0 0 0 0 0 1 0 0 0 0 0 0 0 0 0 0 0]
        [0 0 0 0 0 0 1 0 0 0 0 0 0 0 0 0 0]
        [0 0 0 0 0 0 0 1 0 0 0 0 0 0 0 0 0]
        [0 0 0 0 0 0 0 0 1 0 0 0 0 0 0 0 0]
        [0 0 0 0 0 0 0 0 0 1 0 0 0 0 0 0 0]
        [0 0 0 0 0 0 0 0 0 0 1 0 0 0 0 0 0]
        [0 0 0 0 0 0 0 0 0 0 0 1 0 0 0 0 0]
        [0 0 0 0 0 0 0 0 0 0 0 0 1 0 0 0 0]
        [0 0 0 0 0 0 0 0 0 0 0 0 0 1 0 0 0]
        [0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 0 0]
        [0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 0]
        [0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1]

    The function takes an optional argument, the prime `p` over
    which to work::

        sage: convert_from_milnor_matrix(17, 'adem', 3)                                 # needs sage.combinat sage.modules
        [2 1 1 2]
        [0 2 0 1]
        [1 2 0 0]
        [0 1 0 0]
    """
    mat = convert_to_milnor_matrix(n, basis, p, generic)
    if mat.nrows():
        return convert_to_milnor_matrix(n, basis, p, generic).inverse()
    return mat


@cached_function
def steenrod_algebra_basis(n, basis='milnor', p=2, **kwds):
    r"""
    Basis for the Steenrod algebra in degree `n`.

    INPUT:

    - ``n`` -- nonnegative integer
    - ``basis`` -- string, which basis to use (default: ``'milnor'``)
    - ``p`` -- positive prime number (default: 2)
    - ``profile`` -- profile function (default: ``None``); this
      is just passed on to the functions :func:`milnor_basis` and
      :func:`pst_basis`
    - ``truncation_type`` -- truncation type, either 0 or ``Infinity``
      (default: ``Infinity`` if no profile function is specified,
      0 otherwise).  This is just passed on to the function
      :func:`milnor_basis`.
    - ``generic`` -- boolean (default: ``None``)

    OUTPUT:

    Tuple of objects representing basis elements for the Steenrod algebra
    in dimension n.

    The choices for the string ``basis`` are as follows; see the
    documentation for :mod:`sage.algebras.steenrod.steenrod_algebra`
    for details on each basis:

    - ``'milnor'`` -- Milnor basis
    - ``'serre-cartan'`` or ``'adem'`` or ``'admissible'`` -- Serre-Cartan basis
    - ``'pst'``, ``'pst_rlex'``, ``'pst_llex'``, ``'pst_deg'``, ``'pst_revz'`` --
      various `P^s_t`-bases
    - ``'comm'``, ``'comm_rlex'``, ``'comm_llex'``, ``'comm_deg'``, ``'comm_revz'``, or
      any of these with ``'_long'`` appended -- various commutator bases

    The rest of these bases are only defined when `p=2`.

    - ``'wood_y'`` -- Wood's Y basis
    - ``'wood_z'`` -- Wood's Z basis
    - ``'wall'`` or ``'wall_long'`` -- Wall's basis
    - ``'arnon_a'`` or ``'arnon_a_long'`` -- Arnon's A basis
    - ``'arnon_c'`` -- Arnon's C basis

    EXAMPLES::

        sage: # needs sage.combinat
        sage: from sage.algebras.steenrod.steenrod_algebra_bases import steenrod_algebra_basis
        sage: steenrod_algebra_basis(7, 'milnor') # indirect doctest
        ((0, 0, 1), (1, 2), (4, 1), (7,))
        sage: steenrod_algebra_basis(5)   # milnor basis is the default
        ((2, 1), (5,))

    Bases in negative dimensions are empty::

        sage: steenrod_algebra_basis(-2, 'wall')                                        # needs sage.combinat
        ()

    The third (optional) argument to 'steenrod_algebra_basis' is the
    prime p::

        sage: # needs sage.combinat
        sage: steenrod_algebra_basis(9, 'milnor', p=3)
        (((1,), (1,)), ((0,), (2,)))
        sage: steenrod_algebra_basis(9, 'milnor', 3)
        (((1,), (1,)), ((0,), (2,)))
        sage: steenrod_algebra_basis(17, 'milnor', 3)
        (((2,), ()), ((1,), (3,)), ((0,), (0, 1)), ((0,), (4,)))

    Other bases::

        sage: # needs sage.combinat
        sage: steenrod_algebra_basis(7, 'admissible')
        ((7,), (6, 1), (4, 2, 1), (5, 2))
        sage: steenrod_algebra_basis(13, 'admissible', p=3)
        ((1, 3, 0), (0, 3, 1))
        sage: steenrod_algebra_basis(5, 'wall')
        (((2, 2), (0, 0)), ((1, 1), (1, 0)))
        sage: steenrod_algebra_basis(5, 'wall_long')
        (((2, 2), (0, 0)), ((1, 1), (1, 0)))
        sage: steenrod_algebra_basis(5, 'pst-rlex')
        (((0, 1), (2, 1)), ((1, 1), (0, 2)))
    """
    from .steenrod_algebra_misc import get_basis_name
    try:
        if n < 0 or int(n) != n:
            return ()
    except TypeError:
        return ()

    generic = kwds.get("generic", p != 2)

    basis_name = get_basis_name(basis, p, generic=generic)
    if basis_name.find('long') >= 0:
        basis_name = basis_name.rsplit('_', 1)[0]

    profile = kwds.get("profile", None)
    if (profile is not None and profile != () and profile != ((), ())
            and basis != 'milnor' and basis.find('pst') == -1):
        raise ValueError("profile functions may only be used with the Milnor or pst bases")

    # Milnor basis
    if basis_name == 'milnor':
        return milnor_basis(n, p, **kwds)
    # Serre-Cartan basis
    elif basis_name == 'serre-cartan':
        return serre_cartan_basis(n, p, **kwds)
    # Atomic bases, p odd:
    elif generic and (basis_name.find('pst') >= 0
                    or basis_name.find('comm') >= 0):
        return atomic_basis_odd(n, basis_name, p, **kwds)
    # Atomic bases, p=2
    elif not generic and (basis_name == 'woody' or basis_name == 'woodz'
                     or basis_name == 'wall' or basis_name == 'arnona'
                     or basis_name.find('pst') >= 0
                     or basis_name.find('comm') >= 0):
        return atomic_basis(n, basis_name, **kwds)
    # Arnon 'C' basis
    elif not generic and basis == 'arnonc':
        return arnonC_basis(n)
    else:
        raise ValueError("unknown basis: %s at the prime %s" % (basis, p))


# helper functions for producing bases

def restricted_partitions(n, l, no_repeats=False):
    """
    Iterator over 'restricted' partitions of `n`: partitions with parts taken
    from list `l`.

    INPUT:

    - ``n`` -- nonnegative integer
    - ``l`` -- list of positive integers
    - ``no_repeats`` -- boolean (default: ``False``); if ``True``,
      only return partitions with no repeated parts

    OUTPUT: iterator of lists

    One could also use ``Partitions(n, parts_in=l)``, but this
    function may be faster.  Also, while ``Partitions(n, parts_in=l,
    max_slope=-1)`` should in theory return the partitions of `n` with
    parts in ``l`` with no repetitions, the ``max_slope=-1`` argument
    is ignored, so it does not work.  (At the moment, the
    ``no_repeats=True`` case is the only one used in the code.)

    .. TODO::

        This should be re-implemented in a non-recursive way.

    EXAMPLES::

        sage: from sage.algebras.steenrod.steenrod_algebra_bases import restricted_partitions
        sage: list(restricted_partitions(10, [7,5,1]))
        [[7, 1, 1, 1], [5, 5], [5, 1, 1, 1, 1, 1], [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]]
        sage: list(restricted_partitions(10, [6,5,4,3,2,1], no_repeats=True))
        [[6, 4], [6, 3, 1], [5, 4, 1], [5, 3, 2], [4, 3, 2, 1]]
        sage: list(restricted_partitions(10, [6,4,2]))
        [[6, 4], [6, 2, 2], [4, 4, 2], [4, 2, 2, 2], [2, 2, 2, 2, 2]]
        sage: list(restricted_partitions(10, [6,4,2], no_repeats=True))
        [[6, 4]]

    ``l`` may have repeated elements. If ``no_repeats`` is ``False``, this
    has no effect. If ``no_repeats`` is ``True``, and if the repeated
    elements appear consecutively in ``l``, then each element may be
    used only as many times as it appears in ``l``::

        sage: list(restricted_partitions(10, [6,4,2,2], no_repeats=True))
        [[6, 4], [6, 2, 2]]
        sage: list(restricted_partitions(10, [6,4,2,2,2], no_repeats=True))
        [[6, 4], [6, 2, 2], [4, 2, 2, 2]]

    (If the repeated elements don't appear consecutively, the results
    are likely meaningless, containing several partitions more than
    once, for example.)

    In the following examples, ``no_repeats`` is ``False``::

        sage: list(restricted_partitions(10, [6,4,2]))
        [[6, 4], [6, 2, 2], [4, 4, 2], [4, 2, 2, 2], [2, 2, 2, 2, 2]]
        sage: list(restricted_partitions(10, [6,4,2,2,2]))
        [[6, 4], [6, 2, 2], [4, 4, 2], [4, 2, 2, 2], [2, 2, 2, 2, 2]]
        sage: list(restricted_partitions(10, [6,4,4,4,2,2,2,2,2,2]))
        [[6, 4], [6, 2, 2], [4, 4, 2], [4, 2, 2, 2], [2, 2, 2, 2, 2]]
    """
    if n < 0:
        return

    if n == 0:
        yield []
        return

    index = 1 if no_repeats else 0
    old_i = 0
    for i in l:
        if old_i != i:
            for sigma in restricted_partitions(n - i, l[index:], no_repeats):
                yield [i] + sigma
        index += 1
        old_i = i


def xi_degrees(n, p=2, reverse=True):
    r"""
    Decreasing list of degrees of the `\xi_i`'s, starting in degree `n`.

    INPUT:

    - ``n`` -- integer
    - ``p`` -- prime number (default: 2)
    - ``reverse`` -- boolean (default: ``True``)

    OUTPUT: list of integers

    When `p=2`: decreasing list of the degrees of the `\xi_i`'s with
    degree at most `n`.

    At odd primes: decreasing list of these degrees, each divided by
    `2(p-1)`.

    If ``reverse`` is ``False``, then return an increasing list rather
    than a decreasing one.

    EXAMPLES::

        sage: sage.algebras.steenrod.steenrod_algebra_bases.xi_degrees(17)
        [15, 7, 3, 1]
        sage: sage.algebras.steenrod.steenrod_algebra_bases.xi_degrees(17, reverse=False)
        [1, 3, 7, 15]
        sage: sage.algebras.steenrod.steenrod_algebra_bases.xi_degrees(17, p=3)
        [13, 4, 1]
        sage: sage.algebras.steenrod.steenrod_algebra_bases.xi_degrees(400, p=17)
        [307, 18, 1]
    """
    from sage.rings.integer import Integer
    if n <= 0:
        return []
    N = Integer(n*(p-1) + 1)
    l = [(p**d-1)//(p-1) for d in range(1, N.exact_log(p)+1)]
    if reverse:
        l.reverse()
    return l


########################################################
# Functions for defining bases.

# These should each return a tuple of tuples of the appropriate form
# for the basis.  For example, at the prime 2, the Milnor basis
# element Sq(a,b,c,...) corresponds to the tuple (a, b, c, ...), while
# at odd primes, the element Q_i Q_j ... P(a, b, ...) corresponds to
# the pair ((i, j, ...), (a, b, ...)).  See each function for more
# information.

def milnor_basis(n, p=2, **kwds):
    r"""
    Milnor basis in dimension `n` with profile function ``profile``.

    INPUT:

    - ``n`` -- nonnegative integer

    - ``p`` -- positive prime number (default: 2)

    - ``profile`` -- profile function (default: ``None``).
      Together with ``truncation_type``, specify the profile function
      to be used; ``None`` means the profile function for the entire
      Steenrod algebra.  See
      :mod:`sage.algebras.steenrod.steenrod_algebra` and
      :func:`SteenrodAlgebra <sage.algebras.steenrod.steenrod_algebra.SteenrodAlgebra>`
      for information on profile functions.

    - ``truncation_type`` -- truncation type, either 0 or ``Infinity``
      (default: ``Infinity`` if no profile function is specified,
      0 otherwise)

    OUTPUT: tuple of mod `p` Milnor basis elements in dimension `n`

    At the prime 2, the Milnor basis consists of symbols of the form
    `\text{Sq}(m_1, m_2, ..., m_t)`, where each
    `m_i` is a nonnegative integer and if `t>1`, then
    `m_t \neq 0`. At odd primes, it consists of symbols of the
    form `Q_{e_1} Q_{e_2} ... P(m_1, m_2, ..., m_t)`,
    where `0 \leq e_1 < e_2 < ...`, each `m_i` is a
    nonnegative integer, and if `t>1`, then
    `m_t \neq 0`.

    EXAMPLES::

        sage: # needs sage.combinat
        sage: from sage.algebras.steenrod.steenrod_algebra_bases import milnor_basis
        sage: milnor_basis(7)
        ((0, 0, 1), (1, 2), (4, 1), (7,))
        sage: milnor_basis(7, 2)
        ((0, 0, 1), (1, 2), (4, 1), (7,))
        sage: milnor_basis(4, 2)
        ((1, 1), (4,))
        sage: milnor_basis(4, 2, profile=[2,1])
        ((1, 1),)
        sage: milnor_basis(4, 2, profile=(), truncation_type=0)
        ()
        sage: milnor_basis(4, 2, profile=(), truncation_type=Infinity)
        ((1, 1), (4,))
        sage: milnor_basis(9, 3)
        (((1,), (1,)), ((0,), (2,)))
        sage: milnor_basis(17, 3)
        (((2,), ()), ((1,), (3,)), ((0,), (0, 1)), ((0,), (4,)))
        sage: milnor_basis(48, p=5)
        (((), (0, 1)), ((), (6,)))
        sage: len(milnor_basis(100,3))
        13
        sage: len(milnor_basis(200,7))
        0
        sage: len(milnor_basis(240,7))
        3
        sage: len(milnor_basis(240,7, profile=((),()), truncation_type=Infinity))
        3
        sage: len(milnor_basis(240,7, profile=((),()), truncation_type=0))
        0
    """
    generic = kwds.get('generic', p != 2)

    if n == 0:
        if not generic:
            return ((),)
        else:
            return (((), ()),)

    from sage.rings.infinity import Infinity
    from sage.combinat.integer_vector_weighted import WeightedIntegerVectors
    profile = kwds.get("profile", None)
    trunc = kwds.get("truncation_type", None)
    if trunc is None:
        if profile is not None:
            trunc = 0
        else:
            trunc = Infinity

    result = []
    if not generic:
        for mono in WeightedIntegerVectors(n, xi_degrees(n, reverse=False)):
            exponents = list(mono)
            while exponents and exponents[-1] == 0:
                exponents.pop(-1)
            # check profile:
            okay = True
            if profile is not None and len(profile) > 0:
                for i in range(len(exponents)):
                    if ((len(profile) > i and exponents[i] >= 2**profile[i])
                        or (len(profile) <= i and trunc < Infinity
                            and exponents[i] >= 2**trunc)):
                        okay = False
                        break
            else:
                # profile is empty
                okay = (trunc == Infinity)
            if okay:
                result.append(tuple(exponents))
    else:  # p odd
        # first find the P part of each basis element.
        # in this part of the code (the P part), all dimensions are
        # divided by 2(p-1).
        for dim in range(n//(2*(p-1)) + 1):
            if dim == 0:
                P_result = [[0]]
            else:
                P_result = []
            for mono in WeightedIntegerVectors(dim, xi_degrees(dim, p=p, reverse=False)):
                p_mono = list(mono)
                while p_mono and p_mono[-1] == 0:
                    p_mono.pop(-1)
                if p_mono:
                    P_result.append(p_mono)
            # now find the Q part of the basis element.
            # dimensions here are back to normal.
            for p_mono in P_result:
                deg = n - 2*dim*(p-1)
                q_degrees = [1+2*(p-1)*d for d in
                             xi_degrees(int((deg - 1)//(2*(p-1))), p)] + [1]
                q_degrees_decrease = q_degrees
                q_degrees.reverse()
                if deg % (2*(p-1)) <= len(q_degrees):
                    # if this inequality fails, no way to have a partition
                    # with distinct parts.
                    for sigma in restricted_partitions(deg,
                                                       q_degrees_decrease,
                                                       no_repeats=True):
                        index = 0
                        q_mono = []
                        for q in q_degrees:
                            if q in sigma:
                                q_mono.append(index)
                            index += 1
                        # check profile:
                        okay = True
                        if profile is not None and (len(profile[0]) > 0
                                                    or len(profile[1]) > 0):
                            # check profile function for q_mono
                            for i in q_mono:
                                if ((len(profile[1]) > i and profile[1][i] == 1)
                                    or (len(profile[1]) <= i and trunc == 0)):
                                    okay = False
                                    break
                            # check profile function for p_mono
                            for i in range(len(p_mono)):
                                if okay and ((len(profile[0]) > i and p_mono[i] >= p**profile[0][i])
                                             or (len(profile[0]) <= i and trunc < Infinity
                                                 and p_mono[i] >= p**trunc)):
                                    okay = False
                                    break
                        else:
                            # profile is empty
                            okay = (trunc == Infinity)
                        if okay:
                            if list(p_mono) == [0]:
                                p_mono = []
                            result.append((tuple(q_mono), tuple(p_mono)))
    return tuple(result)


def serre_cartan_basis(n, p=2, bound=1, **kwds):
    r"""
    Serre-Cartan basis in dimension `n`.

    INPUT:

    - ``n`` -- nonnegative integer
    - ``bound`` -- positive integer (optional)
    - ``prime`` -- positive prime number (default: 2)

    OUTPUT: tuple of mod `p` Serre-Cartan basis elements in dimension `n`

    The Serre-Cartan basis consists of 'admissible monomials in the
    Steenrod squares'. Thus at the prime 2, it consists of monomials
    `\text{Sq}^{m_1} \text{Sq}^{m_2} ... \text{Sq}^{m_t}` with `m_i
    \geq 2m_{i+1}` for each `i`. At odd primes, it consists of
    monomials `\beta^{e_0} P^{s_1} \beta^{e_1} P^{s_2} ...  P^{s_k}
    \beta^{e_k}` with each `e_i` either 0 or 1, `s_i \geq p s_{i+1} +
    e_i` for all `i`, and `s_k \geq 1`.

    EXAMPLES::

        sage: from sage.algebras.steenrod.steenrod_algebra_bases import serre_cartan_basis
        sage: serre_cartan_basis(7)
        ((7,), (6, 1), (4, 2, 1), (5, 2))
        sage: serre_cartan_basis(13,3)
        ((1, 3, 0), (0, 3, 1))
        sage: serre_cartan_basis(50,5)
        ((1, 5, 0, 1, 1), (1, 6, 1))

    If optional argument ``bound`` is present, include only those monomials
    whose last term is at least ``bound`` (when `p=2`), or those for which
    `s_k - e_k \geq bound` (when `p` is odd). ::

        sage: serre_cartan_basis(7, bound=2)
        ((7,), (5, 2))
        sage: serre_cartan_basis(13, 3, bound=3)
        ((1, 3, 0),)
    """
    generic = kwds.get('generic', p != 2)

    if n == 0:
        return ((),)
    else:
        if not generic:
            # Build basis recursively.  last = last term.
            # last is >= bound, and we will append (last,) to the end of
            # elements from serre_cartan_basis (n - last, bound=2 * last).
            # This means that 2 last <= n - last, or 3 last <= n.
            result = [(n,)]
            for last in range(bound, 1+n//3):
                for vec in serre_cartan_basis(n - last, bound=2 * last):
                    new = vec + (last,)
                    result.append(new)
        else: # p odd
            if n % (2 * (p-1)) == 0 and n//(2 * (p-1)) >= bound:
                result = [(0, int(n//(2 * (p-1))), 0)]
            elif n == 1:
                result = [(1,)]
            else:
                result = []
            # 2 cases: append P^{last}, or append P^{last} beta
            # case 1: append P^{last}
            for last in range(bound, 1+n//(2*(p - 1))):
                if n - 2*(p-1)*last > 0:
                    for vec in serre_cartan_basis(n - 2*(p-1)*last,
                                                  p, p*last, generic=generic):
                        result.append(vec + (last,0))
            # case 2: append P^{last} beta
            if bound == 1:
                bound = 0
            for last in range(bound+1, 1+n//(2*(p - 1))):
                basis = serre_cartan_basis(n - 2*(p-1)*last - 1,
                                           p, p*last, generic=generic)
                for vec in basis:
                    if vec == ():
                        vec = (0,)
                    new = vec + (last, 1)
                    result.append(new)
    return tuple(result)


def atomic_basis(n, basis, **kwds):
    r"""
    Basis for dimension `n` made of elements in 'atomic' degrees:
    degrees of the form `2^i (2^j - 1)`.

    This works at the prime 2 only.

    INPUT:

    - ``n`` -- nonnegative integer

    - ``basis`` -- string, the name of the basis

    - ``profile`` -- profile function (default: ``None``).
      Together with ``truncation_type``, specify the profile function
      to be used; ``None`` means the profile function for the entire
      Steenrod algebra.  See
      :mod:`sage.algebras.steenrod.steenrod_algebra` and
      :func:`SteenrodAlgebra` for information on profile functions.

    - ``truncation_type`` -- truncation type, either 0 or ``Infinity``
      (default: ``Infinity`` if no profile function is specified,
      0 otherwise).

    OUTPUT: tuple of basis elements in dimension `n`

    The atomic bases include Wood's Y and Z bases, Wall's basis,
    Arnon's A basis, the `P^s_t`-bases, and the commutator
    bases. (All of these bases are constructed similarly, hence their
    constructions have been consolidated into a single function. Also,
    see the documentation for :func:`steenrod_algebra_basis` for
    descriptions of them.)  For `P^s_t`-bases, you may also specify a
    profile function and truncation type; profile functions are ignored
    for the other bases.

    EXAMPLES::

        sage: from sage.algebras.steenrod.steenrod_algebra_bases import atomic_basis
        sage: atomic_basis(6,'woody')
        (((1, 0), (0, 1), (0, 0)), ((2, 0), (1, 0)), ((1, 1),))
        sage: atomic_basis(8,'woodz')
        (((2, 0), (0, 1), (0, 0)), ((0, 2), (0, 0)), ((1, 1), (1, 0)), ((3, 0),))
        sage: atomic_basis(6,'woodz') == atomic_basis(6, 'woody')
        True
        sage: atomic_basis(9,'woodz') == atomic_basis(9, 'woody')
        False

    Wall's basis::

        sage: atomic_basis(8,'wall')
        (((2, 2), (1, 0), (0, 0)), ((2, 0), (0, 0)), ((2, 1), (1, 1)), ((3, 3),))

    Arnon's A basis::

        sage: atomic_basis(7,'arnona')
        (((0, 0), (1, 1), (2, 2)), ((0, 0), (2, 1)), ((1, 0), (2, 2)), ((2, 0),))

    `P^s_t`-bases::

        sage: atomic_basis(7,'pst_rlex')
        (((0, 1), (1, 1), (2, 1)), ((0, 1), (1, 2)), ((2, 1), (0, 2)), ((0, 3),))
        sage: atomic_basis(7,'pst_llex')
        (((0, 1), (1, 1), (2, 1)), ((0, 1), (1, 2)), ((0, 2), (2, 1)), ((0, 3),))
        sage: atomic_basis(7,'pst_deg')
        (((0, 1), (1, 1), (2, 1)), ((0, 1), (1, 2)), ((0, 2), (2, 1)), ((0, 3),))
        sage: atomic_basis(7,'pst_revz')
        (((0, 1), (1, 1), (2, 1)), ((0, 1), (1, 2)), ((0, 2), (2, 1)), ((0, 3),))

    Commutator bases::

        sage: atomic_basis(7,'comm_rlex')
        (((0, 1), (1, 1), (2, 1)), ((0, 1), (1, 2)), ((2, 1), (0, 2)), ((0, 3),))
        sage: atomic_basis(7,'comm_llex')
        (((0, 1), (1, 1), (2, 1)), ((0, 1), (1, 2)), ((0, 2), (2, 1)), ((0, 3),))
        sage: atomic_basis(7,'comm_deg')
        (((0, 1), (1, 1), (2, 1)), ((0, 1), (1, 2)), ((0, 2), (2, 1)), ((0, 3),))
        sage: atomic_basis(7,'comm_revz')
        (((0, 1), (1, 1), (2, 1)), ((0, 1), (1, 2)), ((0, 2), (2, 1)), ((0, 3),))
    """
    def degree_dictionary(n, basis):
        """
        Dictionary of atomic degrees for basis up to degree `n`.

        The keys for the dictionary are the atomic degrees -- the numbers of
        the form `2^i (2^j - 1)` -- which are less than or equal to `n`. The value
        associated to such a degree depends on basis; it has the form
        `(s,t)`, where `(s,t)` is a pair of integers which indexes the
        corresponding element.
        """
        dict = {}
        if basis.find('wood') >= 0:
            k = 0
            m = 0
            deg = 2**m * (2**(k+1) - 1)
            while deg <= n:
                dict[deg] = (m,k)
                if m > 0:
                    m = m - 1
                    k = k + 1
                else:
                    m = k + 1
                    k = 0
                deg = 2**m * (2**(k+1) - 1)
        elif basis.find('wall') >= 0 or basis.find('arnon') >= 0:
            k = 0
            m = 0
            deg = 2**k * (2**(m-k+1) - 1)
            while deg <= n:
                dict[deg] = (m,k)
                if k == 0:
                    m = m + 1
                    k = m
                else:
                    k = k - 1
                deg = 2**k * (2**(m-k+1) - 1)
        elif basis.find('pst') >= 0 or basis.find('comm') >= 0:
            s = 0
            t = 1
            deg = 2**s * (2**t - 1)
            while deg <= n:
                if basis.find('pst') >= 0:
                    dict[deg] = (s,t)
                else:  # comm
                    dict[deg] = (s,t)
                if s == 0:
                    s = t
                    t = 1
                else:
                    s = s - 1
                    t = t + 1
                deg = 2**s * (2**t - 1)
        return dict

    def sorting_pair(s, t, basis):   # pair used for sorting the basis
        if basis.find('wood') >= 0 and basis.find('z') >= 0:
            return (-s-t,-s)
        elif basis.find('wood') >= 0 or basis.find('wall') >= 0 or \
                basis.find('arnon') >= 0:
            return (-s,-t)
        elif basis.find('rlex') >= 0:
            return (t,s)
        elif basis.find('llex') >= 0:
            return (s,t)
        elif basis.find('deg') >= 0:
            return (s+t,t)
        elif basis.find('revz') >= 0:
            return (s+t,s)

    from sage.rings.infinity import Infinity
    profile = kwds.get("profile", None)
    trunc = kwds.get("truncation_type", None)
    if profile is not None and trunc is None:
        trunc = 0

    if n == 0:
        return ((),)
    else:
        result = []
        degrees_etc = degree_dictionary(n, basis)
        degrees = list(degrees_etc)
        for sigma in restricted_partitions(n, degrees, no_repeats=True):
            big_list = [degrees_etc[part] for part in sigma]
            big_list.sort(key=lambda x: sorting_pair(x[0], x[1], basis))
            # reverse = True)
            # arnon: sort like wall, then reverse end result
            if basis.find('arnon') >= 0:
                big_list.reverse()

            # check profile:
            okay = True
            if basis.find('pst') >= 0:
                if profile is not None and len(profile) > 0:
                    for (s,t) in big_list:
                        if ((len(profile) > t-1 and profile[t-1] <= s)
                            or (len(profile) <= t-1 and trunc < Infinity)):
                            okay = False
                            break
            if okay:
                result.append(tuple(big_list))
        return tuple(result)


@cached_function
def arnonC_basis(n, bound=1):
    r"""
    Arnon's C basis in dimension `n`.

    INPUT:

    - ``n`` -- nonnegative integer

    - ``bound`` -- positive integer (optional)

    OUTPUT: tuple of basis elements in dimension `n`

    The elements of Arnon's C basis are monomials of the form
    `\text{Sq}^{t_1} ... \text{Sq}^{t_m}` where for each
    `i`, we have `t_i \leq 2t_{i+1}` and
    `2^i | t_{m-i}`.

    EXAMPLES::

        sage: from sage.algebras.steenrod.steenrod_algebra_bases import arnonC_basis
        sage: arnonC_basis(7)
        ((7,), (2, 5), (4, 3), (4, 2, 1))

    If optional argument ``bound`` is present, include only those monomials
    whose first term is at least as large as ``bound``::

        sage: arnonC_basis(7,3)
        ((7,), (4, 3), (4, 2, 1))
    """
    if n == 0:
        return ((),)

    # Build basis recursively.  first = first term.
    # first is >= bound, and we will prepend (first,) to the
    # elements from arnonC_basis (n - first, first / 2).
    # first also must be divisible by 2**(len(old-basis-elt))
    # This means that 3 first <= 2 n.
    result = [(n,)]
    for first in range(bound, 1 + 2 * n // 3):
        tup = (first,)
        result.extend(tup + vec
                      for vec in arnonC_basis(n - first, max(first // 2, 1))
                      if not first % 2**len(vec))
    return tuple(result)


def atomic_basis_odd(n, basis, p, **kwds):
    r"""
    `P^s_t`-bases and commutator basis in dimension `n` at odd primes.

    This function is called ``atomic_basis_odd`` in analogy with
    :func:`atomic_basis`.

    INPUT:

    - ``n`` -- nonnegative integer

    - ``basis`` -- string, the name of the basis

    - ``p`` -- positive prime number

    - ``profile`` -- profile function (default: ``None``).
      Together with ``truncation_type``, specify the profile function
      to be used; ``None`` means the profile function for the entire
      Steenrod algebra.  See
      :mod:`sage.algebras.steenrod.steenrod_algebra` and
      :func:`SteenrodAlgebra` for information on profile functions.

    - ``truncation_type`` -- truncation type, either 0 or ``Infinity``
      (default: ``Infinity`` if no profile function is specified,
      0 otherwise).

    OUTPUT: tuple of basis elements in dimension `n`

    The only possible difference in the implementations for `P^s_t`
    bases and commutator bases is that the former make sense, and
    require filtering, if there is a nontrivial profile function.
    This function is called by :func:`steenrod_algebra_basis`, and it
    will not be called for commutator bases if there is a profile
    function, so we treat the two bases exactly the same.

    EXAMPLES::

        sage: # needs sage.combinat
        sage: from sage.algebras.steenrod.steenrod_algebra_bases import atomic_basis_odd
        sage: atomic_basis_odd(8, 'pst_rlex', 3)
        (((), (((0, 1), 2),)),)
        sage: atomic_basis_odd(18, 'pst_rlex', 3)
        (((0, 2), ()), ((0, 1), (((1, 1), 1),)))
        sage: atomic_basis_odd(18, 'pst_rlex', 3, profile=((), (2,2,2)))
        (((0, 2), ()),)
    """
    def sorting_pair(s, t, basis):   # pair used for sorting the basis
        if basis.find('rlex') >= 0:
            return (t, s)
        elif basis.find('llex') >= 0:
            return (s, t)
        elif basis.find('deg') >= 0:
            return (s + t, t)
        elif basis.find('revz') >= 0:
            return (s + t, s)

    generic = kwds.get('generic', p != 2)
    if n == 0:
        if not generic:
            return ((),)
        else:
            return (((), ()),)

    from sage.rings.integer import Integer
    from sage.rings.infinity import Infinity
    from sage.combinat.integer_vector_weighted import WeightedIntegerVectors
    profile = kwds.get("profile", None)
    trunc = kwds.get("truncation_type", 0)

    result = []
    for dim in range(n//(2*p-2) + 1):
        P_result = []
        for v in WeightedIntegerVectors(dim, xi_degrees(dim, p=p, reverse=False)):
            mono = []
            for t, a in enumerate(v):
                for s, pow in enumerate(Integer(a).digits(p)):
                    if pow > 0:
                        mono.append(((s, t+1), pow))
            P_result.append(mono)
        for p_mono in P_result:
            p_mono.sort(key=lambda x: sorting_pair(x[0][0], x[0][1], basis))
            deg = n - 2*dim*(p-1)
            q_degrees = [1+2*(p-1)*d for d in
                         xi_degrees((deg - 1)//(2*(p-1)), p)] + [1]
            q_degrees_decrease = q_degrees
            q_degrees.reverse()
            if deg % (2*(p-1)) <= len(q_degrees):
                # if this inequality fails, no way to have a partition
                # with distinct parts.
                for sigma in restricted_partitions(deg,
                                                   q_degrees_decrease,
                                                   no_repeats=True):
                    q_mono = [index for index, q in enumerate(q_degrees)
                              if q in sigma]
                    # check profile:
                    okay = True
                    if profile is not None and profile != ((), ()):
                        # check profile function for q_mono
                        for i in q_mono:
                            if ((len(profile[1]) > i and profile[1][i] == 1)
                                or (len(profile[1]) <= i and trunc == 0)):
                                okay = False
                                break

                        for ((s, t), _) in p_mono:
                            if ((len(profile[0]) > t-1 and profile[0][t-1] <= s)
                                or (len(profile[0]) <= t-1 and trunc < Infinity)):
                                okay = False
                                break

                    if okay:
                        if list(p_mono) == [0]:
                            p_mono = []
                        result.append((tuple(q_mono), tuple(p_mono)))
    return tuple(result)


#############################################################################
def steenrod_basis_error_check(dim, p, **kwds):
    """
    This performs crude error checking.

    INPUT:

    - ``dim`` -- nonnegative integer
    - ``p`` -- positive prime number

    OUTPUT: none

    This checks to see if the different bases have the same length, and
    if the change-of-basis matrices are invertible. If something goes
    wrong, an error message is printed.

    This function checks at the prime `p` as the dimension goes up
    from 0 to ``dim``.

    If you set the Sage verbosity level to a positive integer (using
    ``set_verbose(n)``), then some extra messages will be printed.

    EXAMPLES::

        sage: # long time
        sage: from sage.algebras.steenrod.steenrod_algebra_bases import steenrod_basis_error_check
        sage: steenrod_basis_error_check(15, 2)
        sage: steenrod_basis_error_check(15, 2, generic=True)
        sage: steenrod_basis_error_check(40, 3)
        sage: steenrod_basis_error_check(80, 5)
    """
    from sage.misc.verbose import verbose
    generic = kwds.get('generic', p != 2)

    if not generic:
        bases = ('adem','woody', 'woodz', 'wall', 'arnona', 'arnonc',
                 'pst_rlex', 'pst_llex', 'pst_deg', 'pst_revz',
                 'comm_rlex', 'comm_llex', 'comm_deg', 'comm_revz')
    else:
        bases = ('adem',
                 'pst_rlex', 'pst_llex', 'pst_deg', 'pst_revz',
                 'comm_rlex', 'comm_llex', 'comm_deg', 'comm_revz')

    for i in range(dim):
        if i % 5 == 0:
            verbose("up to dimension %s" % i)
        milnor_dim = len(steenrod_algebra_basis.f(i,'milnor',p=p,generic=generic))
        for B in bases:
            if milnor_dim != len(steenrod_algebra_basis.f(i,B,p,generic=generic)):
                print("problem with milnor/{} in dimension {}".format(B, i))
            mat = convert_to_milnor_matrix.f(i,B,p,generic=generic)
            if mat.nrows() != 0 and not mat.is_invertible():
                print("%s invertibility problem in dim %s at p=%s" % (B, i, p))

    verbose("done checking, no profiles")

    bases = ('pst_rlex', 'pst_llex', 'pst_deg', 'pst_revz')
    if not generic:
        profiles = [(4,3,2,1), (2,2,3,1,1), (0,0,0,2)]
    else:
        profiles = [((3,2,1), ()), ((), (2,1,2)), ((3,2,1), (2,2,2,2))]

    for i in range(dim):
        if i % 5 == 0:
            verbose("up to dimension %s" % i)
        for pro in profiles:
            milnor_dim = len(steenrod_algebra_basis.f(i,'milnor',p=p,profile=pro,generic=generic))
            for B in bases:
                if milnor_dim != len(steenrod_algebra_basis.f(i,B,p,profile=pro,generic=generic)):
                    print("problem with milnor/%s in dimension %s with profile %s" % (B, i, pro))

    verbose("done checking with profiles")



# sage_setup: distribution = sagemath-combinat
r"""
Dyck paths

This is an implementation of the abstract base class
:class:`sage.combinat.path_tableaux.path_tableau.PathTableau`.
This is the simplest implementation of a path tableau and is included
to provide a convenient test case and for pedagogical purposes.

In this implementation we have sequences of nonnegative integers. These
are required to be the heights Dyck words (except that we do not require
the sequence to start or end at height zero). These are in bijection
with skew standard tableaux with at most two rows. Sequences which start
and end at height zero are in bijection with noncrossing perfect matchings.

AUTHORS:

- Bruce Westbury (2018): initial version
"""

#*****************************************************************************
#       Copyright (C) 2018 Bruce Westbury <bruce.westbury@gmail.com>,
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.combinat.path_tableaux.path_tableau import PathTableau, PathTableaux
from sage.combinat.combinatorial_map import combinatorial_map
from sage.combinat.dyck_word import DyckWord
from sage.combinat.perfect_matching import PerfectMatching
from sage.combinat.skew_tableau import SkewTableau
from sage.combinat.tableau import Tableau, StandardTableau
from sage.rings.integer import Integer

###############################################################################


class DyckPath(PathTableau):
    r"""
    An instance is the sequence of nonnegative
    integers given by the heights of a Dyck word.

    INPUT:

    - a sequence of nonnegative integers
    - a two row standard skew tableau
    - a Dyck word
    - a noncrossing perfect matching

    EXAMPLES::

        sage: path_tableaux.DyckPath([0,1,2,1,0])
        [0, 1, 2, 1, 0]

        sage: w = DyckWord([1,1,0,0])
        sage: path_tableaux.DyckPath(w)
        [0, 1, 2, 1, 0]

        sage: p = PerfectMatching([(1,2), (3,4)])
        sage: path_tableaux.DyckPath(p)
        [0, 1, 0, 1, 0]

        sage: t = Tableau([[1,2,4],[3,5,6]])
        sage: path_tableaux.DyckPath(t)
        [0, 1, 2, 1, 2, 1, 0]

        sage: st = SkewTableau([[None, 1,4],[2,3]])
        sage: path_tableaux.DyckPath(st)
        [1, 2, 1, 0, 1]

    Here we illustrate the slogan that promotion = rotation::

        sage: t = path_tableaux.DyckPath([0,1,2,3,2,1,0])
        sage: t.to_perfect_matching()
        [(0, 5), (1, 4), (2, 3)]

        sage: t = t.promotion()
        sage: t.to_perfect_matching()
        [(0, 3), (1, 2), (4, 5)]

        sage: t = t.promotion()
        sage: t.to_perfect_matching()
        [(0, 1), (2, 5), (3, 4)]

        sage: t = t.promotion()
        sage: t.to_perfect_matching()
        [(0, 5), (1, 4), (2, 3)]
    """
    @staticmethod
    def __classcall_private__(cls, ot):
        r"""
        This ensures that a tableau is only ever constructed as an
        ``element_class`` call of an appropriate parent.

        TESTS::

            sage: t = path_tableaux.DyckPath([0,1,2,1,0])

            sage: t.parent()
            <sage.combinat.path_tableaux.dyck_path.DyckPaths_with_category object at ...>
        """
        return DyckPaths()(ot)

    def __init__(self, parent, ot, check=True):
        r"""
        Initialize a Dyck path.

        TESTS::

            sage: D = path_tableaux.DyckPath(Tableau([[1,2], [3,4]]))
            sage: TestSuite(D).run()

            sage: D = path_tableaux.DyckPath(PerfectMatching([(1,4), (2,3), (5,6)]))
            sage: TestSuite(D).run()

            sage: t = path_tableaux.DyckPath([0,1,2,3,2,1,0])
            sage: TestSuite(t).run()

            sage: path_tableaux.DyckPath(PerfectMatching([(1, 3), (2, 4), (5, 6)]))
            Traceback (most recent call last):
            ...
            ValueError: the perfect matching must be non crossing
            sage: path_tableaux.DyckPath(Tableau([[1,2,5],[3,5,6]]))
            Traceback (most recent call last):
            ...
            ValueError: the tableau must be standard
            sage: path_tableaux.DyckPath(Tableau([[1,2,4],[3,5,6],[7]]))
            Traceback (most recent call last):
            ...
            ValueError: the tableau must have at most two rows
            sage: path_tableaux.DyckPath(SkewTableau([[None, 1,4],[2,3],[5]]))
            Traceback (most recent call last):
            ...
            ValueError: the skew tableau must have at most two rows
            sage: path_tableaux.DyckPath([0,1,2.5,1,0])
            Traceback (most recent call last):
            ...
            ValueError: [0, 1, 2.5..., 1, 0] is not a sequence of integers
            sage: path_tableaux.DyckPath(Partition([3,2,1]))
            Traceback (most recent call last):
            ...
            ValueError: invalid input [3, 2, 1]
        """
        w = None

        if isinstance(ot, DyckWord):
            w = ot.heights()

        elif isinstance(ot, PerfectMatching):
            if ot.is_noncrossing():
                u = [1]*ot.size()
                for a in ot.arcs():
                    u[a[1]-1] = 0
                w = DyckWord(u).heights()
            else:
                raise ValueError("the perfect matching must be non crossing")

        elif isinstance(ot, Tableau):
            if len(ot) > 2:
                raise ValueError("the tableau must have at most two rows")
            if ot.is_standard():
                u = [1] * ot.size()
                for i in ot[1]:
                    u[i-1] = 0
                w = DyckWord(u).heights()
            else:
                raise ValueError("the tableau must be standard")

        elif isinstance(ot, SkewTableau):
            if len(ot) > 2:
                raise ValueError("the skew tableau must have at most two rows")
            # The check that ot is standard is not implemented
            c = ot.to_chain()
            w = [0]*len(c)
            for i,a in enumerate(c):
                if len(a) == 1:
                    w[i] = a[0]
                else:
                    w[i] = a[0] - a[1]

        elif isinstance(ot, (list, tuple)):
            try:
                w = tuple([Integer(a) for a in ot])
            except TypeError:
                raise ValueError("%s is not a sequence of integers" % ot)

        if w is None:
            raise ValueError("invalid input %s" % ot)

        PathTableau.__init__(self, parent, w, check=check)

    def check(self):
        """
        Check that ``self`` is a valid path.

        EXAMPLES::

            sage: path_tableaux.DyckPath([0,1,0,-1,0]) # indirect doctest
            Traceback (most recent call last):
            ...
            ValueError: [0, 1, 0, -1, 0] has a negative entry

            sage: path_tableaux.DyckPath([0,1,3,1,0]) # indirect doctest
            Traceback (most recent call last):
            ...
            ValueError: [0, 1, 3, 1, 0] is not a Dyck path
        """
        if any(a < 0 for a in self):
            raise ValueError("%s has a negative entry" % (str(self)))
        for i in range(len(self) - 1):
            if abs(self[i + 1] - self[i]) != 1:
                raise ValueError("%s is not a Dyck path" % str(self))

    def local_rule(self, i):
        r"""
        This has input a list of objects. This method first takes
        the list of objects of length three consisting of the `(i-1)`-st,
        `i`-th and `(i+1)`-term and applies the rule. It then replaces
        the `i`-th object  by the object returned by the rule.

        EXAMPLES::

            sage: t = path_tableaux.DyckPath([0,1,2,3,2,1,0])
            sage: t.local_rule(3)
            [0, 1, 2, 1, 2, 1, 0]

        TESTS::

            sage: t = path_tableaux.DyckPath([0,1,2,3,2,1,0])
            sage: t.local_rule(0)
            Traceback (most recent call last):
            ...
            ValueError: 0 is not a valid integer
            sage: t.local_rule(5)
            [0, 1, 2, 3, 2, 1, 0]
            sage: t.local_rule(6)
            Traceback (most recent call last):
            ...
            ValueError: 6 is not a valid integer
        """
        def _rule(x):
            """
            This is the rule on a sequence of three letters.
            """
            return abs(x[0]-x[1]+x[2])

        if not (i > 0 and i < len(self)-1):
            raise ValueError("%d is not a valid integer" % i)

        with self.clone() as result:
            result[i] = _rule(self[i-1:i+2])

        return result

    def is_skew(self):
        r"""
        Return ``True`` if ``self`` is skew and ``False`` if not.

        EXAMPLES::

            sage: path_tableaux.DyckPath([0,1,2,1]).is_skew()
            False

            sage: path_tableaux.DyckPath([1,0,1,2,1]).is_skew()
            True
        """
        return self[0] != 0

    @combinatorial_map(name='to Dyck word')
    def to_DyckWord(self):
        r"""
        Convert ``self`` to a Dyck word.

        EXAMPLES::

            sage: c = path_tableaux.DyckPath([0,1,2,1,0])
            sage: c.to_DyckWord()
            [1, 1, 0, 0]
        """
        return DyckWord(heights_sequence=list(self))

    def descents(self):
        r"""
        Return the descent set of ``self``.

        EXAMPLES::

            sage: path_tableaux.DyckPath([0,1,2,1,2,1,0,1,0]).descents()
            {3, 6}
        """
        result = set()

        for i in range(1,len(self)-1):
            if self[i] < self[i-1] and self[i] < self[i+1]:
                result.add(i)

        return result

    def to_word(self):
        r"""
        Return the word in the alphabet `\{0,1\}` associated to ``self``.

        EXAMPLES::

            sage: path_tableaux.DyckPath([1,0,1,2,1]).to_word()
            [0, 1, 1, 0]
        """
        return [(self[i+1] - self[i] + 1) // 2 for i in range(self.size()-1)]

    def to_perfect_matching(self):
        r"""
        Return the perfect matching associated to ``self``.

        EXAMPLES::

            sage: path_tableaux.DyckPath([0,1,2,1,2,1,0,1,0]).to_perfect_matching()
            [(0, 5), (1, 2), (3, 4), (6, 7)]

        TESTS::

            sage: path_tableaux.DyckPath([1,2,1,2,1,0,1]).to_perfect_matching()
            Traceback (most recent call last):
            ...
            ValueError: [1, 2, 1, 2, 1, 0, 1] does not start at 0
        """
        if self.is_skew():
            raise ValueError( "%s does not start at 0" % (str(self)) )
        w = self.to_word()
        y = DyckWord(w)
        pairs = set()
        for i, a in enumerate(y):
            c = y.associated_parenthesis(i)
            if i < c:
                pairs.add((i,c))
        return PerfectMatching(pairs)

    def to_tableau(self):
        r"""
        Return the skew tableau associated to ``self``.

        EXAMPLES::

            sage: T = path_tableaux.DyckPath([0,1,2,3,2,3])
            sage: T.to_tableau()
            [[1, 2, 3, 5], [4]]

            sage: U = path_tableaux.DyckPath([2,3,2,3])
            sage: U.to_tableau()
            [[None, None, 1, 3], [2]]
        """
        w = self.to_word()
        top = [i + 1 for i, a in enumerate(w) if a == 1]
        bot = [i + 1 for i, a in enumerate(w) if a == 0]
        if self.is_skew():
            return SkewTableau([[None]*self[0]+top, bot])
        else:
            return StandardTableau([top, bot])


class DyckPaths(PathTableaux):
    """
    The parent class for DyckPath.
    """

    def _an_element_(self):
        """
        Return an element of ``self``.

        EXAMPLES::

            sage: path_tableaux.DyckPaths()._an_element_()
            [0, 1, 2, 1, 0]
        """
        return DyckPath([0,1,2,1,0])

    Element = DyckPath



# sage_setup: distribution = sagemath-modules
"""
The Steenrod algebra
"""
from sage.misc.lazy_import import lazy_import
lazy_import('sage.algebras.steenrod.steenrod_algebra', ['SteenrodAlgebra', 'Sq'])
del lazy_import



# sage_setup: distribution = sagemath-combinat
r"""
Subwords

A subword of a word `w` is a word obtained by deleting the letters at some
(non necessarily adjacent) positions in `w`. It is not to be confused with the
notion of factor where one keeps adjacent positions in `w`. Sometimes it is
useful to allow repeated uses of the same letter of `w` in a "generalized"
subword. We call this a subword with repetitions.

For example:

- "bnjr" is a subword of the word "bonjour" but not a factor;

- "njo" is both a factor and a subword of the word "bonjour";

- "nr" is a subword of "bonjour";

- "rn" is not a subword of "bonjour";

- "nnu" is not a subword of "bonjour";

- "nnu" is a subword with repetitions of "bonjour";

A word can be given either as a string, as a list or as a tuple.

As repetition can occur in the initial word, in general subwords
of a given word form an enumerated multiset rather than a set!

.. TODO::

    - implement subwords with repetitions

    - implement the category of ``EnumeratedMultiset`` and inheritate from
      it when needed (i.e. the initial word has repeated letters)

AUTHORS:

- Mike Hansen: initial version

- Florent Hivert (2009/02/06): doc improvements + new methods + bug fixes
"""
# ****************************************************************************
#       Copyright (C) 2007 Mike Hansen <mhansen@gmail.com>,
#                     2014 Vincent Delecroix <20100.delecroix@gmail.com>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from __future__ import annotations
from collections.abc import Iterator
import itertools

from sage.structure.parent import Parent

from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets

import sage.misc.prandom as prandom
from sage.rings.integer import Integer
from sage.sets.finite_enumerated_set import FiniteEnumeratedSet


def _stringification(data):
    r"""
    TESTS::

        sage: from sage.combinat.subword import _stringification
        sage: _stringification(['a','b','c'])
        'abc'
    """
    return ''.join(data)


def Subwords(w, k=None, element_constructor=None):
    """
    Return the set of subwords of ``w``.

    INPUT:

    - ``w`` -- a word (can be a list, a string, a tuple or a word)

    - ``k`` -- an optional integer to specify the length of subwords

    - ``element_constructor`` -- an optional function that will be used
      to build the subwords

    EXAMPLES::

        sage: S = Subwords(['a','b','c']); S
        Subwords of ['a', 'b', 'c']
        sage: S.first()
        []
        sage: S.last()
        ['a', 'b', 'c']
        sage: S.list()
        [[], ['a'], ['b'], ['c'], ['a', 'b'], ['a', 'c'], ['b', 'c'], ['a', 'b', 'c']]

    The same example using string, tuple or a word::

        sage: S = Subwords('abc'); S
        Subwords of 'abc'
        sage: S.list()
        ['', 'a', 'b', 'c', 'ab', 'ac', 'bc', 'abc']

        sage: S = Subwords((1,2,3)); S
        Subwords of (1, 2, 3)
        sage: S.list()
        [(), (1,), (2,), (3,), (1, 2), (1, 3), (2, 3), (1, 2, 3)]

        sage: w = Word([1,2,3])
        sage: S = Subwords(w); S
        Subwords of word: 123
        sage: S.list()
        [word: , word: 1, word: 2, word: 3, word: 12, word: 13, word: 23, word: 123]

    Using word with specified length::

        sage: S = Subwords(['a','b','c'], 2); S
        Subwords of ['a', 'b', 'c'] of length 2
        sage: S.list()
        [['a', 'b'], ['a', 'c'], ['b', 'c']]

    An example that uses the ``element_constructor`` argument::

        sage: p = Permutation([3,2,1])
        sage: Subwords(p, element_constructor=tuple).list()
        [(), (3,), (2,), (1,), (3, 2), (3, 1), (2, 1), (3, 2, 1)]
        sage: Subwords(p, 2, element_constructor=tuple).list()
        [(3, 2), (3, 1), (2, 1)]
    """
    if element_constructor is None:
        datatype = type(w)  # 'datatype' is the type of w
        if datatype is list or datatype is tuple:
            element_constructor = datatype
        elif datatype is str:
            element_constructor = _stringification
        else:
            from sage.combinat.words.words import Words
            try:
                alphabet = w.parent().alphabet()
                element_constructor = Words(alphabet)
            except AttributeError:
                element_constructor = list

    if k is None:
        return Subwords_w(w, element_constructor)
    if not isinstance(k, (int, Integer)):
        raise ValueError("k should be an integer")
    if k < 0 or k > len(w):
        return FiniteEnumeratedSet([])
    return Subwords_wk(w, k, element_constructor)


class Subwords_w(Parent):
    r"""
    Subwords of a given word.
    """

    def __init__(self, w, element_constructor):
        """
        TESTS::

            sage: TestSuite(Subwords([1,2,3])).run()
            sage: TestSuite(Subwords('sage')).run()
        """
        Parent.__init__(self, category=FiniteEnumeratedSets())
        self._w = w
        self._build = element_constructor

    def __eq__(self, other) -> bool:
        r"""
        Equality test.

        TESTS::

            sage: Subwords([1,2,3]) == Subwords([1,2,3])
            True
            sage: Subwords([1,2,3]) == Subwords([1,3,2])
            False
        """
        return self.__class__ == other.__class__ and self._w == other._w and self._build == other._build

    def __ne__(self, other) -> bool:
        r"""
        TESTS::

            sage: Subwords([1,2,3]) != Subwords([1,2,3])
            False
            sage: Subwords([1,2,3]) != Subwords([1,3,2])
            True
        """
        return not self == other

    def __reduce__(self):
        r"""
        Pickle (how to construct back the object).

        TESTS::

            sage: S = Subwords((1,2,3))
            sage: S == loads(dumps(S))
            True
            sage: S = Subwords('123')
            sage: S == loads(dumps(S))
            True
            sage: S = Subwords(('a',(1,2,3),('a','b'),'ir'))
            sage: S == loads(dumps(S))
            True
        """
        return (Subwords_w, (self._w, self._build))

    def _repr_(self) -> str:
        """
        TESTS::

            sage: repr(Subwords([1,2,3])) # indirect doctest
            'Subwords of [1, 2, 3]'
        """
        return "Subwords of {!r}".format(self._w)

    def __contains__(self, w) -> bool:
        """
        TESTS::

            sage: [] in Subwords([1,2,3,4,3,4,4])
            True
            sage: [2,3,3,4] in Subwords([1,2,3,4,3,4,4])
            True
            sage: [5,5,3] in Subwords([1,3,3,5,4,5,3,5])
            True
            sage: [3,5,5,3] in Subwords([1,3,3,5,4,5,3,5])
            True
            sage: [3,5,5,3,4] in Subwords([1,3,3,5,4,5,3,5])
            False
            sage: [2,3,3,4] in Subwords([1,2,3,4,3,4,4])
            True
            sage: [2,3,3,1] in Subwords([1,2,3,4,3,4,4])
            False
        """
        return smallest_positions(self._w, w) is not False

    def cardinality(self) -> Integer:
        """
        EXAMPLES::

            sage: Subwords([1,2,3]).cardinality()
            8
        """
        return Integer(1) << len(self._w)

    def first(self):
        """
        EXAMPLES::

            sage: Subwords([1,2,3]).first()
            []
            sage: Subwords((1,2,3)).first()
            ()
            sage: Subwords('123').first()
            ''
        """
        return self._build([])

    def last(self):
        """
        EXAMPLES::

            sage: Subwords([1,2,3]).last()
            [1, 2, 3]
            sage: Subwords((1,2,3)).last()
            (1, 2, 3)
            sage: Subwords('123').last()
            '123'
        """
        return self._build(self._w)

    def random_element(self):
        r"""
        Return a random subword with uniform law.

        EXAMPLES::

            sage: S1 = Subwords([1,2,3,2,1,3])
            sage: S2 = Subwords([4,6,6,6,7,4,5,5])
            sage: for i in range(100):
            ....:   w = S1.random_element()
            ....:   if w in S2:
            ....:       assert not w
            sage: for i in range(100):
            ....:   w = S2.random_element()
            ....:   if w in S1:
            ....:       assert not w
        """
        return self._build(elt for elt in self._w if prandom.randint(0, 1))

    def __iter__(self) -> Iterator:
        r"""
        EXAMPLES::

            sage: Subwords([1,2,3]).list()
            [[], [1], [2], [3], [1, 2], [1, 3], [2, 3], [1, 2, 3]]
            sage: Subwords((1,2,3)).list()
            [(), (1,), (2,), (3,), (1, 2), (1, 3), (2, 3), (1, 2, 3)]
            sage: Subwords('123').list()
            ['', '1', '2', '3', '12', '13', '23', '123']
        """
        return itertools.chain(*[Subwords_wk(self._w, i, self._build)
                                 for i in range(len(self._w) + 1)])


class Subwords_wk(Subwords_w):
    r"""
    Subwords with fixed length of a given word.
    """

    def __init__(self, w, k, element_constructor):
        """
        TESTS::

            sage: S = Subwords([1,2,3],2)
            sage: S == loads(dumps(S))
            True
            sage: TestSuite(S).run()
        """
        Subwords_w.__init__(self, w, element_constructor)
        self._k = k

    def __eq__(self, other) -> bool:
        r"""
        Equality test.

        TESTS::

            sage: Subwords([1,2,3],2) == Subwords([1,2,3],2)
            True
            sage: Subwords([1,2,3],2) == Subwords([1,3,2],2)
            False
            sage: Subwords([1,2,3],2) == Subwords([1,2,3],3)
            False
        """
        return Subwords_w.__eq__(self, other) and self._k == other._k

    def __ne__(self, other) -> bool:
        r"""
        TESTS::

            sage: Subwords([1,2,3], 2) != Subwords([1,2,3], 2)
            False
            sage: Subwords([1,2,3], 2) != Subwords([1,3,2], 1)
            True
        """
        return not self == other

    def __reduce__(self):
        r"""
        Pickle (how to construct back the object).

        TESTS::

            sage: S = Subwords('abc',2)
            sage: S == loads(dumps(S))
            True
            sage: S = Subwords(('a',1,'45',(1,2)))
            sage: S == loads(dumps(S))
            True
        """
        return (Subwords_wk, (self._w, self._k, self._build))

    def _repr_(self) -> str:
        """
        TESTS::

            sage: repr(Subwords([1,2,3],2))  # indirect doctest
            'Subwords of [1, 2, 3] of length 2'
        """
        return "{} of length {}".format(Subwords_w._repr_(self), self._k)

    def __contains__(self, w) -> bool:
        """
        TESTS::

            sage: [] in Subwords([1, 3, 3, 5, 4, 5, 3, 5],0)
            True
            sage: [2,3,3,4] in Subwords([1,2,3,4,3,4,4],4)
            True
            sage: [2,3,3,4] in Subwords([1,2,3,4,3,4,4],3)
            False
            sage: [5,5,3] in Subwords([1,3,3,5,4,5,3,5],3)
            True
            sage: [5,5,3] in Subwords([1,3,3,5,4,5,3,5],4)
            False
        """
        return len(w) == self._k and Subwords_w.__contains__(self, w)

    def cardinality(self) -> Integer:
        r"""
        Return the number of subwords of w of length k.

        EXAMPLES::

            sage: Subwords([1,2,3], 2).cardinality()
            3
        """
        return Integer(len(self._w)).binomial(self._k)

    def first(self):
        r"""
        EXAMPLES::

            sage: Subwords([1,2,3],2).first()
            [1, 2]
            sage: Subwords([1,2,3],0).first()
            []
            sage: Subwords((1,2,3),2).first()
            (1, 2)
            sage: Subwords((1,2,3),0).first()
            ()
            sage: Subwords('123',2).first()
            '12'
            sage: Subwords('123',0).first()
            ''
        """
        return self._build(self._w[i] for i in range(self._k))

    def last(self):
        r"""
        EXAMPLES::

            sage: Subwords([1,2,3],2).last()
            [2, 3]
            sage: Subwords([1,2,3],0).last()
            []
            sage: Subwords((1,2,3),2).last()
            (2, 3)
            sage: Subwords((1,2,3),0).last()
            ()
            sage: Subwords('123',2).last()
            '23'
            sage: Subwords('123',0).last()
            ''

        TESTS::

            sage: Subwords('123', 0).last()  # trac 10534
            ''
        """
        n = len(self._w)
        return self._build(self._w[i] for i in range(n - self._k, n))

    def random_element(self):
        r"""
        Return a random subword of given length with uniform law.

        EXAMPLES::

            sage: S1 = Subwords([1,2,3,2,1],3)
            sage: S2 = Subwords([4,4,5,5,4,5,4,4],3)
            sage: for i in range(100):
            ....:   w = S1.random_element()
            ....:   if w in S2:
            ....:       assert not w
            sage: for i in range(100):
            ....:   w = S2.random_element()
            ....:   if w in S1:
            ....:       assert not w
        """
        sample = prandom.sample(self._w, self._k)
        if self._build is list:
            return sample
        return self._build(sample)

    def __iter__(self) -> Iterator:
        """
        EXAMPLES::

            sage: Subwords([1,2,3],2).list()
            [[1, 2], [1, 3], [2, 3]]
            sage: Subwords([1,2,3],0).list()
            [[]]
            sage: Subwords((1,2,3),2).list()
            [(1, 2), (1, 3), (2, 3)]
            sage: Subwords((1,2,3),0).list()
            [()]
            sage: Subwords('abc',2).list()
            ['ab', 'ac', 'bc']
            sage: Subwords('abc',0).list()
            ['']
        """
        if self._k > len(self._w):
            return iter([])
        iterator = itertools.combinations(self._w, self._k)
        if self._build is tuple:
            return iterator
        return (self._build(x) for x in iterator)


def smallest_positions(word, subword, pos=0) -> list | bool:
    """
    Return the smallest positions for which ``subword`` appears as a
    subword of ``word``.

    If ``pos`` is specified, then it returns the positions
    of the first appearance of subword starting at ``pos``.

    If ``subword`` is not found in ``word``, then return ``False``.

    EXAMPLES::

        sage: sage.combinat.subword.smallest_positions([1,2,3,4], [2,4])
        [1, 3]
        sage: sage.combinat.subword.smallest_positions([1,2,3,4,4], [2,4])
        [1, 3]
        sage: sage.combinat.subword.smallest_positions([1,2,3,3,4,4], [3,4])
        [2, 4]
        sage: sage.combinat.subword.smallest_positions([1,2,3,3,4,4], [3,4],2)
        [2, 4]
        sage: sage.combinat.subword.smallest_positions([1,2,3,3,4,4], [3,4],3)
        [3, 4]
        sage: sage.combinat.subword.smallest_positions([1,2,3,4], [2,3])
        [1, 2]
        sage: sage.combinat.subword.smallest_positions([1,2,3,4], [5,5])
        False
        sage: sage.combinat.subword.smallest_positions([1,3,3,4,5],[3,5])
        [1, 4]
        sage: sage.combinat.subword.smallest_positions([1,3,3,5,4,5,3,5],[3,5,3])
        [1, 3, 6]
        sage: sage.combinat.subword.smallest_positions([1,3,3,5,4,5,3,5],[3,5,3],2)
        [2, 3, 6]
        sage: sage.combinat.subword.smallest_positions([1,2,3,4,3,4,4],[2,3,3,1])
        False
        sage: sage.combinat.subword.smallest_positions([1,3,3,5,4,5,3,5],[3,5,3],3)
        False

    TESTS:

    We check for :issue:`5534`::

        sage: w = ["a", "b", "c", "d"]; ww = ["b", "d"]
        sage: x = sage.combinat.subword.smallest_positions(w, ww); ww
        ['b', 'd']
    """
    pos -= 1
    n = len(word)
    res = [None] * len(subword)
    for i, swi in enumerate(subword):
        for j in range(pos + 1, n + 1):
            if j == n:
                return False
            if word[j] == swi:
                pos = j
                break
        if pos != j:
            return False
        res[i] = pos

    return res



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.graphs sage.modules
r"""
Polyhedral realization of `B(\infty)`
"""

# ***************************************************************************
#       Copyright (C) 2015 Travis Scrimshaw <tscrim at ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
# ***************************************************************************

from sage.structure.parent import Parent
from sage.categories.infinite_enumerated_sets import InfiniteEnumeratedSets
from sage.categories.highest_weight_crystals import HighestWeightCrystals
from sage.combinat.crystals.tensor_product import TensorProductOfCrystals, \
    TensorProductOfCrystalsElement
from sage.combinat.crystals.elementary_crystals import ElementaryCrystal
from sage.combinat.root_system.cartan_type import CartanType


class InfinityCrystalAsPolyhedralRealization(TensorProductOfCrystals):
    r"""
    The polyhedral realization of `B(\infty)`.

    .. NOTE::

        Here we are using anti-Kashiwara notation and might differ from
        some of the literature.

    Consider a Kac-Moody algebra `\mathfrak{g}` of Cartan type `X` with
    index set `I`,  and consider a finite sequence `J = (j_1, j_2, \ldots, j_m)`
    whose support equals `I`.  We extend this to an infinite sequence
    by taking `\bar{J} = J \cdot J \cdot J \cdots`, where `\cdot` denotes
    concatenation of sequences. Let

    .. MATH::

        B_J = B_{j_m} \otimes \cdots \otimes B_{j_2} \otimes B_{j_1},

    where `B_i` is an
    :class:`~sage.combinat.crystals.elementary_crystals.ElementaryCrystal`.

    As given in Theorem 2.1.1 of [Ka1993]_, there exists a strict crystal embedding
    `\Psi_i \colon B(\infty) \to B_i \otimes B(\infty)` defined by `u_{\infty}
    \mapsto b_i(0) \otimes u_{\infty}`, where `b_i(0) \in B_i` and `u_{\infty}`
    is the (unique) highest weight element in `B(\infty)`. This is sometimes
    known as the *Kashiwara embedding* [NZ1997]_ (though, in [NZ1997]_, the target
    of this map is denoted by `\ZZ_J^\infty`).  By iterating this embedding by
    taking `\Psi_J = \Psi_{j_n} \circ \Psi_{j_{n-1}} \circ \cdots \circ
    \Psi_{j_1}`, we obtain the following strict crystal embedding:

    .. MATH::

        \Psi_J^n \colon B(\infty) \to B_J^{\otimes n} \otimes B(\infty).

    We note there is a natural analog of Lemma 10.6.2 in [HK2002]_ that
    for any `b \in B(\infty)`, there exists a positive integer `N` such that

    .. MATH::

        \Psi^N_J(b) = \left( \bigotimes_{k=1}^N b^{(k)} \right)
        \otimes u_{\infty}.

    Therefore we can model elements `b \in B(\infty)` by considering
    an infinite list of elements `b^{(k)} \in B_J` and defining the crystal
    structure by:

    .. MATH::

        \begin{aligned}
        \mathrm{wt}(b) & = \sum_{k=1}^N \mathrm{wt}(b^{(k)})
        \\ e_i(b) & = e_i\left( \left( \bigotimes_{k=1}^N b^{(k)} \right)
        \right) \otimes u_{\infty},
        \\ f_i(b) & = f_i\left( \left( \bigotimes_{k=1}^N b^{(k)} \right)
        \right) \otimes u_{\infty},
        \\ \varepsilon_i(b) & = \max_{ e_i^k(b) \neq 0 } k,
        \\ \varphi_i(b) & = \varepsilon_i(b) - \langle \mathrm{wt}(b),
        h_i^{\vee} \rangle.
        \end{aligned}

    To translate this into a finite list, we consider a finite sequence
    `b_1 \otimes \cdots \otimes b_N` and if

    .. MATH::

        f_i\left( b^{(1)} \otimes \cdots b^{(N-1)} \otimes b^{(N)} \right)
        = b^{(1)} \otimes \cdots \otimes b^{(N-1)} \otimes
        f_i\left( b^{(N)} \right),

    then we take the image as `b^{(1)} \otimes \cdots \otimes f_i\left(
    b^{(N)} \right) \otimes b^{(N+1)}`. Similarly we remove `b^{(N)}` if
    we have `b^{(N)} = \bigotimes_{k=1}^m b_{j_k}(0)`. Additionally if

    .. MATH::

        e_i\left( b^{(1)} \otimes \cdots \otimes b^{(N-1)} \otimes
        b^{(N)} \right) = b^{(1)} \otimes \cdots \otimes b^{(N-1)}
        \otimes e_i\left( b^{(N)} \right),

    then we consider this to be `0`.

    INPUT:

    - ``cartan_type`` -- a Cartan type
    - ``seq`` -- (default: ``None``) a finite sequence whose support
      equals the index set of the Cartan type; if ``None``, then this
      is the index set

    EXAMPLES::

        sage: B = crystals.infinity.PolyhedralRealization(['A',2])
        sage: mg = B.module_generators[0]; mg
        [0, 0]
        sage: mg.f_string([2,1,2,2])
        [0, -3, -1, 0, 0, 0]

    An example of type `B_2`::

        sage: B = crystals.infinity.PolyhedralRealization(['B',2])
        sage: mg = B.module_generators[0]; mg
        [0, 0]
        sage: mg.f_string([2,1,2,2])
        [0, -2, -1, -1, 0, 0]

    An example of type `G_2`::

        sage: B = crystals.infinity.PolyhedralRealization(['G',2])
        sage: mg = B.module_generators[0]; mg
        [0, 0]
        sage: mg.f_string([2,1,2,2])
        [0, -3, -1, 0, 0, 0]
    """
    @staticmethod
    def __classcall_private__(cls, cartan_type, seq=None):
        """
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: B1 = crystals.infinity.PolyhedralRealization(['A',2])
            sage: B2 = crystals.infinity.PolyhedralRealization(['A',2], [1,2])
            sage: B1 is B2
            True
        """
        cartan_type = CartanType(cartan_type)
        if seq is None:
            seq = cartan_type.index_set()
        else:
            seq = tuple(seq)
        if set(seq) != set(cartan_type.index_set()):
            raise ValueError("the support of seq is not the index set")
        return super().__classcall__(cls, cartan_type, seq)

    def __init__(self, cartan_type, seq):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: B = crystals.infinity.PolyhedralRealization(['A',2])
            sage: TestSuite(B).run() # long time
        """
        cat = (HighestWeightCrystals(), InfiniteEnumeratedSets())
        Parent.__init__(self, category=cat)
        self._cartan_type = cartan_type
        self._seq = seq
        # These are the additional factors we add as necessary
        self._factors = tuple([ElementaryCrystal(cartan_type, i) for i in seq])
        # public for TensorProductOfCrystals
        self.crystals = self._factors
        self._tp = [C.module_generators[0] for C in self.crystals]
        self.module_generators = (self.element_class(self, self._tp),)

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: crystals.infinity.PolyhedralRealization(['A',2])
            Polyhedral realization of B(oo) of type ['A', 2] using (1, 2)
        """
        return "Polyhedral realization of B(oo) of type {} using {}".format(self._cartan_type, self._seq)

    def finite_tensor_product(self, k):
        """
        Return the finite tensor product of crystals of length ``k``
        by truncating ``self``.

        EXAMPLES::

            sage: B = crystals.infinity.PolyhedralRealization(['A',2])
            sage: B.finite_tensor_product(5)
            Full tensor product of the crystals
             [The 1-elementary crystal of type ['A', 2],
              The 2-elementary crystal of type ['A', 2],
              The 1-elementary crystal of type ['A', 2],
              The 2-elementary crystal of type ['A', 2],
              The 1-elementary crystal of type ['A', 2]]
        """
        N = len(self._factors)
        crystals = [self._factors[i % N] for i in range(k)]
        return TensorProductOfCrystals(*crystals)

    class Element(TensorProductOfCrystalsElement):
        r"""
        An element in the polyhedral realization of `B(\infty)`.
        """
        # For simplicity (and safety), we use the regular crystals implementation

        def epsilon(self, i):
            r"""
            Return `\varepsilon_i` of ``self``.

            EXAMPLES::

                sage: B = crystals.infinity.PolyhedralRealization(['A',2,1])
                sage: mg = B.module_generators[0]
                sage: [mg.epsilon(i) for i in B.index_set()]
                [0, 0, 0]
                sage: elt = mg.f(0)
                sage: [elt.epsilon(i) for i in B.index_set()]
                [1, 0, 0]
                sage: elt = mg.f_string([0,1,2])
                sage: [elt.epsilon(i) for i in B.index_set()]
                [0, 0, 1]
                sage: elt = mg.f_string([0,1,2,2])
                sage: [elt.epsilon(i) for i in B.index_set()]
                [0, 0, 2]
            """
            x = self.e(i)
            eps = 0
            while x is not None:
                x = x.e(i)
                eps = eps + 1
            return eps

        def phi(self, i):
            r"""
            Return `\varphi_i` of ``self``.

            EXAMPLES::

                sage: B = crystals.infinity.PolyhedralRealization(['A',2,1])
                sage: mg = B.module_generators[0]
                sage: [mg.phi(i) for i in B.index_set()]
                [0, 0, 0]
                sage: elt = mg.f(0)
                sage: [elt.phi(i) for i in B.index_set()]
                [-1, 1, 1]
                sage: elt = mg.f_string([0,1])
                sage: [elt.phi(i) for i in B.index_set()]
                [-1, 0, 2]
                sage: elt = mg.f_string([0,1,2,2])
                sage: [elt.phi(i) for i in B.index_set()]
                [1, 1, 0]
            """
            P = self.parent().weight_lattice_realization()
            h = P.simple_coroots()
            omega = P(self.weight()).scalar(h[i])
            return self.epsilon(i) + omega

        def e(self, i):
            """
            Return the action of `e_i` on ``self``.

            EXAMPLES::

                sage: B = crystals.infinity.PolyhedralRealization(['A',2])
                sage: mg = B.module_generators[0]
                sage: all(mg.e(i) is None for i in B.index_set())
                True
                sage: mg.f(1).e(1) == mg
                True
            """
            N = len(self) + 1
            pos = None
            for k in range(1, N):
                if all(self._sig(i,k) > self._sig(i,j) for j in range(1, k)) and \
                        all(self._sig(i,k) >= self._sig(i,j) for j in range(k+1, N)):
                    crystal = self[-k].e(i)
                    pos = k
                    break

            nf = len(self.parent()._factors)
            if pos is None or pos <= nf:
                return None

            l = list(self)
            l[-pos] = crystal
            if pos <= 2*nf and all(b._m == 0 for b in l[-2*nf:-nf]):
                return self.__class__(self.parent(), l[:-nf])
            return self.__class__(self.parent(), l)

        def f(self, i):
            """
            Return the action of `f_i` on ``self``.

            EXAMPLES::

                sage: B = crystals.infinity.PolyhedralRealization(['A',2])
                sage: mg = B.module_generators[0]
                sage: mg.f(1)
                [-1, 0, 0, 0]
                sage: mg.f_string([1,2,2,1])
                [-1, -2, -1, 0, 0, 0]
            """
            N = len(self) + 1
            pos = None
            for k in range(1, N):
                if all(self._sig(i,k) >= self._sig(i,j) for j in range(1, k)) and \
                        all(self._sig(i,k) > self._sig(i,j) for j in range(k+1, N)):
                    crystal = self[-k].f(i)
                    pos = k
                    break

            nf = len(self.parent()._factors)
            if pos <= nf:
                l = list(self)
                l[-pos] = l[-pos].f(i)
                return self.__class__(self.parent(), l + self.parent()._tp)
            return self._set_index(-pos, crystal)

        def truncate(self, k=None):
            r"""
            Truncate ``self`` to have length ``k`` and return as an element
            in a (finite) tensor product of crystals.

            INPUT:

            - ``k`` -- (optional) the length of the truncation; if not
              specified, then returns one more than the current non-ground-state
              elements (i.e. the current list in ``self``)

            EXAMPLES::

                sage: B = crystals.infinity.PolyhedralRealization(['A',2])
                sage: mg = B.module_generators[0]
                sage: elt = mg.f_string([1,2,2,1]); elt
                [-1, -2, -1, 0, 0, 0]
                sage: t = elt.truncate(); t
                [-1, -2, -1, 0, 0, 0]
                sage: t.parent() is B.finite_tensor_product(6)
                True
                sage: elt.truncate(2)
                [-1, -2]
                sage: elt.truncate(10)
                [-1, -2, -1, 0, 0, 0, 0, 0, 0, 0]
            """
            if k is None:
                k = len(self)

            P = self.parent().finite_tensor_product(k)
            if k <= len(self):
                l = self[:k]
            else:
                l = list(self)
                N = len(self.parent()._tp)
                while len(l) < k:
                    i = len(l) % N
                    l.append(self.parent()._tp[i])
            return P(*l)



# sage_setup: distribution = sagemath-combinat
cdef extern from 'symmetrica/def.h':
    INT strict_to_odd_part(OP s, OP o)
    INT odd_to_strict_part(OP o, OP s)
    INT q_core(OP part, OP d,  OP core)
    INT gupta_nm(OP n, OP m, OP res)
    INT gupta_tafel(OP max, OP res)
    INT random_partition(OP nx, OP res)


def strict_to_odd_part_symmetrica(part):
    """
    Implement the bijection between strict partitions
    and partitions with odd parts. input is a VECTOR type partition, the
    result is a partition of the same weight with only odd parts.
    """

    #Make sure that the partition is strict
    cdef INT i
    for i from 0 <= i < len(part)-1:
        if part[i] == part[i+1]:
            raise ValueError("the partition part (= %s) must be strict" % str(part))

    cdef OP cpart, cres
    anfang()
    cpart = callocobject()
    cres = callocobject()

    _op_partition(part, cpart)

    strict_to_odd_part(cpart, cres)

    res = _py(cres)

    freeall(cpart)
    freeall(cres)
    ende()

    return res


def odd_to_strict_part_symmetrica(part):
    """
    Implement the bijection between partitions with odd parts
    and strict partitions. input is a VECTOR type partition, the
    result is a partition of the same weight with different parts.
    """

    #Make sure that the partition is strict
    cdef INT i
    for i from 0 <= i < len(part):
        if part[i] % 2 == 0:
            raise ValueError("the partition part (= %s) must be odd" % str(part))

    cdef OP cpart, cres
    anfang()
    cpart = callocobject()
    cres = callocobject()

    _op_partition(part, cpart)

    odd_to_strict_part(cpart, cres)

    res = _py(cres)

    freeall(cpart)
    freeall(cres)
    ende()

    return res


def q_core_symmetrica(part, d):
    """
    Compute the q-core of a PARTITION object
    part. This is the remaining partition (=res) after
    removing of all hooks of length d (= INTEGER object).
    The result may be an empty object, if the whole
    partition disappears.
    """


    cdef OP cpart, cres, cd
    anfang()
    cpart = callocobject()
    cd = callocobject()
    cres = callocobject()

    _op_partition(part, cpart)
    _op_integer(d, cd)

    q_core(cpart, cd, cres)

    res = _py(cres)

    freeall(cpart)
    freeall(cres)
    freeall(cd)
    ende()

    return res


def gupta_nm_symmetrica(n, m):
    """
    This routine computes the number of partitions
    of n with maximal part m. The result is erg. The
    input n,m must be INTEGER objects. The result is
    freed first to an empty object. The result must
    be a different from m and n.
    """


    cdef OP cn, cm, cres
    anfang()
    cm = callocobject()
    cn = callocobject()
    cres = callocobject()


    _op_integer(n, cn)
    _op_integer(m, cm)

    gupta_nm(cn, cm, cres)

    res = _py(cres)

    freeall(cn)
    freeall(cres)
    freeall(cm)
    ende()

    return res


def gupta_tafel_symmetrica(max):
    """
    It computes the table of the above values. The entry
    n,m is the result of gupta_nm. mat is freed first.
    max must be an INTEGER object, it is the maximum
    weight for the partitions. max must be different from
    result.
    """


    cdef OP cmax, cres
    anfang()

    cmax = callocobject()
    cres = callocobject()


    _op_integer(max, cmax)

    gupta_tafel(cmax, cres)

    res = _py(cres)

    freeall(cmax)
    freeall(cres)
    ende()

    return res


def random_partition_symmetrica(n):
    """
    Return a random partition p of the entered weight w.

    w must be an INTEGER object, p becomes a PARTITION object.
    Type of partition is VECTOR . It uses the algorithm of
    Nijenhuis and Wilf, p.76
    """


    cdef OP cn, cres
    anfang()

    cn = callocobject()
    cres = callocobject()


    _op_integer(n, cn)

    random_partition(cn, cres)

    res = _py(cres)

    freeall(cn)
    freeall(cres)
    ende()

    return res



# sage_setup: distribution = sagemath-combinat
r"""
Ribbon shaped tableaux
"""
# ****************************************************************************
#       Copyright (C) 2007 Mike Hansen <mhansen@gmail.com>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.combinat.skew_tableau import SkewTableau, SkewTableaux, StandardSkewTableaux
from sage.combinat.tableau import Tableaux
from sage.combinat.permutation import descents_composition_first, descents_composition_list, descents_composition_last
from sage.rings.integer import Integer
from sage.categories.infinite_enumerated_sets import InfiniteEnumeratedSets
from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets
from sage.categories.sets_cat import Sets


class RibbonShapedTableau(SkewTableau):
    r"""
    A ribbon shaped tableau.

    For the purposes of this class, a ribbon shaped tableau is a skew
    tableau whose shape is a skew partition which:

    - has at least one cell in row `1`;

    - has at least one cell in column `1`;

    - has exactly one cell in each of `q` consecutive diagonals, for
      some nonnegative integer `q`.

    A ribbon is given by a list of the rows from top to bottom.

    EXAMPLES::

        sage: x = RibbonShapedTableau([[None, None, None, 2, 3], [None, 1, 4, 5], [3, 2]]); x
        [[None, None, None, 2, 3], [None, 1, 4, 5], [3, 2]]
        sage: x.pp()
          .  .  .  2  3
          .  1  4  5
          3  2
        sage: x.shape()
        [5, 4, 2] / [3, 1]

    The entries labeled by ``None`` correspond to the inner partition.
    Using ``None`` is optional; the entries will be shifted accordingly.  ::

        sage: x = RibbonShapedTableau([[2,3],[1,4,5],[3,2]]); x.pp()
          .  .  .  2  3
          .  1  4  5
          3  2

    TESTS::

        sage: r = RibbonShapedTableau([[1], [2,3], [4, 5, 6]])
        sage: r.to_permutation()
        [4, 5, 6, 2, 3, 1]

        sage: RibbonShapedTableau([[1,2],[3,4]]).evaluation()
        [1, 1, 1, 1]
    """
    @staticmethod
    def __classcall_private__(cls, rows):
        r"""
        Return a ribbon shaped tableau object.

        EXAMPLES::

            sage: RibbonShapedTableau([[2,3],[1,4,5]])
            [[None, None, 2, 3], [1, 4, 5]]

        TESTS::

            sage: RibbonShapedTableau([4,5])
            Traceback (most recent call last):
            ...
            TypeError: rows must be lists of positive integers

            sage: RibbonShapedTableau([[2,3],[-4,5]])
            Traceback (most recent call last):
            ...
            TypeError: r must be a list of positive integers
        """
        try:
            r = [tuple(r) for r in rows]
        except TypeError:
            raise TypeError("rows must be lists of positive integers")
        if not r:
            return StandardRibbonShapedTableaux()(r)
        if all(j is None or (isinstance(j, (int, Integer)) and j > 0)
               for i in r for j in i):
            return StandardRibbonShapedTableaux()(r)
        raise TypeError("r must be a list of positive integers")

    def __init__(self, parent, t):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: R = RibbonShapedTableau([[2,3],[1,4,5]])
            sage: TestSuite(R).run()
        """
        if not isinstance(t, SkewTableau):
            # scrubbing None
            t = [[i for i in row if i is not None] for row in t]

            st = []
            space_count = 0
            for row in reversed(t):
                st.append([None] * space_count + row)
                space_count += len(row) - 1
            st.reverse()
            t = st
        else:
            t = list(t)

        SkewTableau.__init__(self, parent, t)

    def height(self):
        """
        Return the height of ``self``.

        The height is given by the number of rows in the outer partition.

        EXAMPLES::

            sage: RibbonShapedTableau([[2,3],[1,4,5]]).height()
            2
        """
        return len(self.outer_shape())

    def spin(self):
        """
        Return the spin of ``self``.

        EXAMPLES::

            sage: RibbonShapedTableau([[2,3],[1,4,5]]).spin()
            1/2
        """
        return Integer(self.height() - 1) / 2

    def width(self):
        """
        Return the width of the ribbon.

        This is given by the length of the longest row in the outer partition.

        EXAMPLES::

            sage: RibbonShapedTableau([[2,3],[1,4,5]]).width()
            4
            sage: RibbonShapedTableau([]).width()
            0
        """
        return len(self[0]) if self else 0


class RibbonShapedTableaux(SkewTableaux):
    """
    The set of all ribbon shaped tableaux.
    """
    @staticmethod
    def __classcall_private__(cls, shape=None, **kwds):
        """
        Normalize input to ensure a unique representation and pick the correct
        class based on input.

        The ``shape`` parameter is currently ignored.

        EXAMPLES::

            sage: S1 = RibbonShapedTableaux([4, 2, 2, 1])
            sage: S2 = RibbonShapedTableaux((4, 2, 2, 1))
            sage: S1 is S2
            True
        """
        # if shape is not None:
        #     from sage.combinat.partition import Partition
        #     return RibbonShapedTableaux_shape(Partition(shape))

        # Otherwise arg0 takes the place of the category in pickling
        return super().__classcall__(cls, **kwds)

    def __init__(self, category=None):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: S = RibbonShapedTableaux()                                            # needs sage.graphs
            sage: TestSuite(S).run()                                                    # needs sage.graphs
        """
        if category is None:
            category = Sets()

        SkewTableaux.__init__(self, category=category)

    def _repr_(self):
        """
        TESTS::

            sage: repr(RibbonShapedTableaux())    # indirect doctest
            'Ribbon shaped tableaux'
        """
        return "Ribbon shaped tableaux"

    Element = RibbonShapedTableau
    options = Tableaux.options

    def from_shape_and_word(self, shape, word):
        """
        Return the ribbon corresponding to the given ribbon shape and word.

        EXAMPLES::

            sage: RibbonShapedTableaux().from_shape_and_word([1,3],[1,3,3,7])
            [[None, None, 1], [3, 3, 7]]
        """
        pos = 0
        r = []
        for l in shape:
            r.append(word[pos:pos + l])
            pos += l
        return self.element_class(self, r)


class StandardRibbonShapedTableaux(StandardSkewTableaux):
    """
    The set of all standard ribbon shaped tableaux.

    INPUT:

    - ``shape`` -- (optional) the composition shape of the rows
    """
    @staticmethod
    def __classcall_private__(cls, shape=None, **kwds):
        """
        Normalize input to ensure a unique representation and pick the correct
        class based on input.

        EXAMPLES::

            sage: S1 = StandardRibbonShapedTableaux([4, 2, 2, 1])
            sage: S2 = StandardRibbonShapedTableaux((4, 2, 2, 1))
            sage: S1 is S2
            True
        """
        if shape is not None:
            from sage.combinat.partition import Partition
            return StandardRibbonShapedTableaux_shape(Partition(shape))

        # Otherwise arg0 takes the place of the category in pickling
        return super().__classcall__(cls, **kwds)

    def __init__(self, category=None):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: # needs sage.graphs sage.modules
            sage: S = StandardRibbonShapedTableaux()
            sage: TestSuite(S).run()
        """
        if category is None:
            category = InfiniteEnumeratedSets()

        StandardSkewTableaux.__init__(self, category=category)

    def _repr_(self):
        """
        TESTS::

            sage: repr(StandardRibbonShapedTableaux())    # indirect doctest
            'Standard ribbon shaped tableaux'
        """
        return "Standard ribbon shaped tableaux"

    def __iter__(self):
        """
        Iterate through ``self``.

        EXAMPLES::

            sage: # needs sage.graphs sage.modules
            sage: it = StandardRibbonShapedTableaux().__iter__()
            sage: [next(it) for x in range(10)]
            [[],
             [[1]],
             [[1, 2]],
             [[1], [2]],
             [[1, 2, 3]],
             [[None, 1], [2, 3]],
             [[None, 2], [1, 3]],
             [[1], [2], [3]],
             [[1, 2, 3, 4]],
             [[None, None, 1], [2, 3, 4]]]
        """
        from sage.combinat.partition import _Partitions
        for p in _Partitions:
            for r in StandardRibbonShapedTableaux_shape(p):
                yield self.element_class(self, r)

    Element = RibbonShapedTableau
    options = Tableaux.options

    def from_shape_and_word(self, shape, word):
        """
        Return the ribbon corresponding to the given ribbon shape and word.

        EXAMPLES::

            sage: StandardRibbonShapedTableaux().from_shape_and_word([2,3],[1,2,3,4,5])
            [[None, None, 1, 2], [3, 4, 5]]
        """
        pos = 0
        r = []
        for l in shape:
            r.append(word[pos:pos + l])
            pos += l
        return self.element_class(self, r)

    def from_permutation(self, p):
        """
        Return a standard ribbon of size ``len(p)`` from a permutation ``p``. The
        lengths of each row are given by the distance between the descents
        of the permutation ``p``.

        EXAMPLES::

            sage: import sage.combinat.ribbon_shaped_tableau as rst
            sage: [StandardRibbonShapedTableaux().from_permutation(p)
            ....:  for p in Permutations(3)]
            [[[1, 2, 3]],
             [[None, 2], [1, 3]],
             [[1, 3], [2]],
             [[None, 1], [2, 3]],
             [[1, 2], [3]],
             [[1], [2], [3]]]
        """
        if not p:
            return self.element_class(self, [])

        comp = p.descents()

        if not comp:
            return self.element_class(self, [p[:]])

        r = [[p[j] for j in range(comp[0])]]
        r.extend([p[j] for j in range(comp[i], comp[i + 1])]
                 for i in range(len(comp) - 1))
        r.append([p[j] for j in range(comp[-1], len(p))])
        r.reverse()
        return self.element_class(self, r)


class StandardRibbonShapedTableaux_shape(StandardRibbonShapedTableaux):
    """
    Class of standard ribbon shaped tableaux of ribbon shape ``shape``.

    EXAMPLES::

        sage: StandardRibbonShapedTableaux([2,2])
        Standard ribbon shaped tableaux of shape [2, 2]
        sage: StandardRibbonShapedTableaux([2,2]).first()
        [[None, 2, 4], [1, 3]]
        sage: StandardRibbonShapedTableaux([2,2]).last()
        [[None, 1, 2], [3, 4]]

        sage: # needs sage.graphs sage.modules
        sage: StandardRibbonShapedTableaux([2,2]).cardinality()
        5
        sage: StandardRibbonShapedTableaux([2,2]).list()
        [[[None, 1, 3], [2, 4]],
         [[None, 1, 2], [3, 4]],
         [[None, 2, 3], [1, 4]],
         [[None, 2, 4], [1, 3]],
         [[None, 1, 4], [2, 3]]]
        sage: StandardRibbonShapedTableaux([3,2,2]).cardinality()
        155
    """
    @staticmethod
    def __classcall_private__(cls, shape):
        """
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: S = StandardRibbonShapedTableaux([2,2])
            sage: S2 = StandardRibbonShapedTableaux((2,2))
            sage: S is S2
            True
        """
        return super(StandardRibbonShapedTableaux, cls).__classcall__(cls, tuple(shape))

    def __init__(self, shape):
        """
        TESTS::

            sage: S = StandardRibbonShapedTableaux([2,2])
            sage: TestSuite(S).run()                                                    # needs sage.graphs sage.modules
        """
        self.shape = shape
        StandardRibbonShapedTableaux.__init__(self, FiniteEnumeratedSets())

    def _repr_(self):
        """
        TESTS::

            sage: StandardRibbonShapedTableaux([2,2])
            Standard ribbon shaped tableaux of shape [2, 2]
        """
        return "Standard ribbon shaped tableaux of shape %s" % list(self.shape)

    def first(self):
        """
        Return the first standard ribbon of ``self``.

        EXAMPLES::

            sage: StandardRibbonShapedTableaux([2,2]).first()
            [[None, 2, 4], [1, 3]]
        """
        return self.from_permutation(descents_composition_first(self.shape))

    def last(self):
        """
        Return the last standard ribbon of ``self``.

        EXAMPLES::

            sage: StandardRibbonShapedTableaux([2,2]).last()
            [[None, 1, 2], [3, 4]]
        """
        return self.from_permutation(descents_composition_last(self.shape))

    def __iter__(self):
        """
        An iterator for the standard ribbon of ``self``.

        EXAMPLES::

            sage: [t for t in StandardRibbonShapedTableaux([2,2])]                      # needs sage.graphs sage.modules
            [[[None, 1, 3], [2, 4]],
             [[None, 1, 2], [3, 4]],
             [[None, 2, 3], [1, 4]],
             [[None, 2, 4], [1, 3]],
             [[None, 1, 4], [2, 3]]]
        """
        for p in descents_composition_list(self.shape):
            yield self.from_permutation(p)


class Ribbon_class(RibbonShapedTableau):
    """
    This exists solely for unpickling ``Ribbon_class`` objects.
    """

    def __setstate__(self, state):
        r"""
        Unpickle old ``Ribbon_class`` objects.

        EXAMPLES::

            sage: loads(b'x\x9ck`J.NLO\xd5K\xce\xcfM\xca\xccK,\xd1+\xcaLJ\xca\xcf\xe3\n\x02S\xf1\xc99\x89\xc5\xc5\\\x85\x8c\x9a\x8d\x85L\xb5\x85\xcc\x1a\xa1\xac\xf1\x19\x89\xc5\x19\x85,~@VNfqI!kl!\x9bFl!\xbb\x06\xc4\x9c\xa2\xcc\xbc\xf4b\xbd\xcc\xbc\x92\xd4\xf4\xd4"\xae\xdc\xc4\xec\xd4x\x18\xa7\x90#\x94\xd1\xb05\xa8\x903\x03\xc80\x022\xb8Rc\x0b\xb95@<c \x8f\x07\xc40\x012xSSK\x93\xf4\x00l\x811\x17')
            [[None, 1, 2], [3, 4]]
            sage: loads(dumps( RibbonShapedTableau([[3,2,1], [1,1]]) ))  # indirect doctest
            [[None, 3, 2, 1], [1, 1]]
        """
        self.__class__ = RibbonShapedTableau
        self.__init__(RibbonShapedTableaux(), state['_list'])


from sage.misc.persist import register_unpickle_override
register_unpickle_override('sage.combinat.ribbon', 'Ribbon_class', Ribbon_class)
register_unpickle_override('sage.combinat.ribbon', 'StandardRibbons_shape', StandardRibbonShapedTableaux)



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.modules

cdef extern from 'symmetrica/def.h':
    INT mult_schubert_schubert(OP a, OP b, OP result)
    INT m_perm_sch(OP a, OP b)
    INT t_SCHUBERT_POLYNOM(OP a, OP b)
    INT t_POLYNOM_SCHUBERT(OP a, OP b)
    INT mult_schubert_variable(OP a, OP i, OP r)
    INT divdiff_perm_schubert(OP perm, OP sb, OP res)
    INT scalarproduct_schubert(OP a, OP b, OP c)
    INT divdiff_schubert(OP a, OP schub, OP res)

    INT t_2SCHUBERT_POLYNOM(OP a,OP b)
    INT mult_schubert_polynom(OP a,OP b,OP c)


cdef object _check_schubert(object a, OP ca):
    if a in Permutations():
        if isinstance(a, builtinlist):
            a = Permutation(a)
        _op_schubert_perm(a, ca)
        return max(a.reduced_word()+[0])
    elif isinstance(a, SchubertPolynomial_class):
        br = a.parent().base_ring()
        if (br == QQ or br == ZZ):
            _op_schubert_sp(a, ca)
            return min(max(i.reduced_word(), default=0) for i in a.support())
        else:
            raise ValueError("a must be a Schubert polynomial over ZZ or QQ")
    else:
        raise TypeError("a must be a permutation or a Schubert polynomial")


def mult_schubert_schubert_symmetrica(a, b):
    """
    Multiply the Schubert polynomials `a` and `b`.

    EXAMPLES::

        sage: symmetrica.mult_schubert_schubert([3,2,1], [3,2,1])
        X[5, 3, 1, 2, 4]
    """
    late_import()

    cdef OP ca = callocobject(), cb = callocobject(), cres = callocobject()

    try:
        max_a = _check_schubert(a, ca)
        max_b = _check_schubert(b, cb)
    except (ValueError, TypeError), err:
        freeall(ca)
        freeall(cb)
        freeall(cres)
        raise err

    sig_on()
    mult_schubert_schubert(ca, cb, cres)
    sig_off()

    res = _py(cres)

    freeall(ca)
    freeall(cb)
    freeall(cres)

    return res


def t_SCHUBERT_POLYNOM_symmetrica(a):
    """
    Convert a Schubert polynomial to a 'regular' multivariate
    polynomial.

    EXAMPLES::

        sage: symmetrica.t_SCHUBERT_POLYNOM([3,2,1])
        x0^2*x1
    """
    late_import()

    cdef OP ca = callocobject(), cres = callocobject()

    try:
        max_a = _check_schubert(a, ca)
    except (ValueError, TypeError), err:
        freeall(ca)
        freeall(cres)
        raise err

    sig_on()
    t_SCHUBERT_POLYNOM(ca, cres)
    sig_off()

    res = _py(cres)

    freeall(ca)
    freeall(cres)

    return res


def t_POLYNOM_SCHUBERT_symmetrica(a):
    """
    Convert a multivariate polynomial a to a Schubert polynomial.

    EXAMPLES::

        sage: R.<x1,x2,x3> = QQ[]
        sage: w0 = x1^2*x2
        sage: symmetrica.t_POLYNOM_SCHUBERT(w0)
        X[3, 2, 1]
    """
    late_import()

    cdef OP ca = callocobject(), cres = callocobject()

    if not isinstance(a, MPolynomial):
        freeall(ca)
        freeall(cres)
        raise TypeError("a (= %s) must be a multivariate polynomial")
    else:
        br = a.parent().base_ring()
        if br != QQ and br != ZZ:
            freeall(ca)
            freeall(cres)
            raise ValueError("a's base ring must be either ZZ or QQ")
        else:
            _op_polynom(a, ca)

    sig_on()
    t_POLYNOM_SCHUBERT(ca, cres)
    sig_off()

    res = _py(cres)

    freeall(ca)
    freeall(cres)

    return res


def mult_schubert_variable_symmetrica(a, i):
    """
    Return the product of `a` and `x_i`.  Note that indexing with `i`
    starts at 1.

    EXAMPLES::

        sage: symmetrica.mult_schubert_variable([3,2,1], 2)
        X[3, 2, 4, 1]
        sage: symmetrica.mult_schubert_variable([3,2,1], 4)
        X[3, 2, 1, 4, 6, 5] - X[3, 2, 1, 5, 4]
    """
    late_import()

    cdef OP ca = callocobject(), ci = callocobject(),  cres = callocobject()

    try:
        max_a = _check_schubert(a, ca)
    except (ValueError, TypeError), err:
        freeall(ca)
        freeall(ci)
        freeall(cres)
        raise err

    _op_integer(i, ci)

    sig_on()
    mult_schubert_variable(ca, ci, cres)
    sig_off()

    res = _py(cres)

    freeall(ca)
    freeall(ci)
    freeall(cres)

    return res


def divdiff_perm_schubert_symmetrica(perm, a):
    r"""
    Return the result of applying the divided difference operator
    `\delta_i` to `a` where `a` is either a permutation or a
    Schubert polynomial over `\QQ`.

    EXAMPLES::

       sage: symmetrica.divdiff_perm_schubert([2,3,1], [3,2,1])
       X[2, 1]
       sage: symmetrica.divdiff_perm_schubert([3,1,2], [3,2,1])
       X[1, 3, 2]
       sage: symmetrica.divdiff_perm_schubert([3,2,4,1], [3,2,1])
       Traceback (most recent call last):
       ...
       ValueError: cannot apply \delta_{[3, 2, 4, 1]} to a (= [3, 2, 1])
    """
    late_import()

    cdef OP ca = callocobject(), cperm = callocobject(),  cres = callocobject()

    try:
        max_a = _check_schubert(a, ca)
    except (ValueError, TypeError), err:
        freeall(ca)
        freeall(cperm)
        freeall(cres)
        raise err

    if perm not in Permutations():
        freeall(ca)
        freeall(cperm)
        freeall(cres)
        raise TypeError("perm must be a permutation")
    else:
        perm = Permutation(perm)
        rw = perm.reduced_word()
        max_perm = max(rw)
        _op_permutation(perm, cperm)

    if max_perm > max_a:
        freeall(ca)
        freeall(cperm)
        freeall(cres)
        raise ValueError(r"cannot apply \delta_{%s} to a (= %s)" % (perm, a))

    sig_on()
    divdiff_perm_schubert(cperm, ca, cres)
    sig_off()

    res = _py(cres)

    freeall(ca)
    freeall(cperm)
    freeall(cres)

    return res


def scalarproduct_schubert_symmetrica(a, b):
    """
    EXAMPLES::

        sage: symmetrica.scalarproduct_schubert([3,2,1], [3,2,1])
        X[1, 3, 5, 2, 4]
        sage: symmetrica.scalarproduct_schubert([3,2,1], [2,1,3])
        X[1, 2, 4, 3]
    """
    late_import()

    cdef OP ca = callocobject(), cb = callocobject(), cres = callocobject()

    try:
        max_a = _check_schubert(a, ca)
        max_b = _check_schubert(b, cb)
    except (ValueError, TypeError), err:
        freeall(ca)
        freeall(cb)
        freeall(cres)
        raise err

    sig_on()
    scalarproduct_schubert(ca, cb, cres)
    sig_off()

    if empty_listp(cres):
        res = Integer(0)
    else:
        res = _py(cres)

    freeall(ca)
    freeall(cb)
    freeall(cres)

    return res


def divdiff_schubert_symmetrica(i, a):
    r"""
    Return the result of applying the divided difference operator
    `\delta_i` to `a` where `a` is either a permutation or a
    Schubert polynomial over `\QQ`.

    EXAMPLES::

       sage: symmetrica.divdiff_schubert(1, [3,2,1])
       X[2, 3, 1]
       sage: symmetrica.divdiff_schubert(2, [3,2,1])
       X[3, 1, 2]
       sage: symmetrica.divdiff_schubert(3, [3,2,1])
       Traceback (most recent call last):
       ...
       ValueError: cannot apply \delta_{3} to a (= [3, 2, 1])
    """
    late_import()

    cdef OP ca = callocobject(), ci = callocobject(),  cres = callocobject()

    try:
        max_a = _check_schubert(a, ca)
    except (ValueError, TypeError), err:
        freeall(ca)
        freeall(ci)
        freeall(cres)
        raise err

    if not isinstance(i, (int, Integer)):
        freeall(ca)
        freeall(ci)
        freeall(cres)
        raise TypeError("i must be an integer")
    else:
        _op_integer(i, ci)

    if i > max_a or i <= 0:
        freeall(ca)
        freeall(ci)
        freeall(cres)
        raise ValueError(r"cannot apply \delta_{%s} to a (= %s)" % (i, a))

    sig_on()
    divdiff_schubert(ci, ca, cres)
    sig_off()

    res = _py(cres)

    freeall(ca)
    freeall(ci)
    freeall(cres)

    return res



# sage_setup: distribution = sagemath-combinat
"""
SAT-Solvers via DIMACS Files

Sage supports calling SAT solvers using the popular DIMACS format. This module implements
infrastructure to make it easy to add new such interfaces and some example interfaces.

Currently, interfaces to **RSat** and **Glucose** are included by default.

.. NOTE::

    Our SAT solver interfaces are 1-based, i.e., literals start at 1. This is consistent with the
    popular DIMACS format for SAT solving but not with Pythion's 0-based convention. However, this
    also allows to construct clauses using simple integers.

AUTHORS:

- Martin Albrecht (2012): first version
- Sébastien Labbé (2018): adding Glucose SAT solver
- Sébastien Labbé (2023): adding Kissat SAT solver

Classes and Methods
-------------------
"""
##############################################################################
#  Copyright (C) 2012 Martin Albrecht <martinralbrecht@googlemail.com>
#  Distributed under the terms of the GNU General Public License (GPL)
#  The full text of the GPL is available at:
#                  https://www.gnu.org/licenses/
##############################################################################

from pathlib import Path
import sys
import subprocess
import shlex
from time import sleep

from sage.features.sat import Glucose as Glucose_executable, Kissat as Kissat_executable
from sage.sat.solvers.satsolver import SatSolver
from sage.misc.temporary_file import tmp_filename


class DIMACS(SatSolver):
    """
    Generic DIMACS Solver.

    .. NOTE::

        Usually, users will not have to use this class directly but some
        class which inherits from this class.

    .. automethod:: __init__
    .. automethod:: __call__
    """

    command = ""

    def __init__(self, command=None, filename=None, verbosity=0, **kwds):
        """
        Construct a new generic DIMACS solver.

        INPUT:

        - ``command`` -- a named format string or list of strings with the
          command to run. The string must contain {input} and may contain
          {output} if the solvers writes the solution to an output
          file. For example "sat-solver {input}" is a valid
          command. If ``None`` then the class variable ``command`` is
          used. (default: ``None``)

        - ``filename`` -- a filename to write clauses to in DIMACS
          format, must be writable. If ``None`` a temporary filename
          is chosen automatically. (default: ``None``)

        - ``verbosity`` -- a verbosity level, where zero means silent
          and anything else means verbose output. (default: ``0``)

        - ``**kwds`` -- accepted for compatibility with other solvers; ignored

        TESTS::

            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: DIMACS()
            DIMACS Solver: ''
        """
        self._headname_file_created_during_init = False
        if filename is None:
            filename = tmp_filename()
            self._headname_file_created_during_init = True

        self._headname = filename
        self._verbosity = verbosity

        if command is not None:
            self._command = command
        else:
            self._command = self.__class__.command

        self._tail = open(tmp_filename(), 'w')
        self._var = 0
        self._lit = 0

    def __repr__(self):
        """
        TESTS::

            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: DIMACS(command="iliketurtles {input}")
            DIMACS Solver: 'iliketurtles {input}'
        """
        command = self._command
        if not isinstance(command, str):
            command = ' '.join((arg if isinstance(arg, str) else arg.executable)
                               for arg in command)
        return "DIMACS Solver: '%s'" % command

    def __del__(self):
        """
        TESTS::

            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: d = DIMACS(command="iliketurtles {input}")
            sage: del d

        We check that files created during initialization are properly
        deleted (:issue:`38328`)::

            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: d = DIMACS(command="iliketurtles {input}")
            sage: filename = d._headname
            sage: os.path.exists(filename)
            True
            sage: del d
            sage: os.path.exists(filename)
            False

        ::

            sage: fn = tmp_filename()
            sage: d = DIMACS(filename=fn)
            sage: del d
        """
        if not self._tail.closed:
            self._tail.close()
        Path(self._tail.name).unlink(missing_ok=True)
        if self._headname_file_created_during_init:
            Path(self._headname).unlink(missing_ok=True)

    def var(self, decision=None):
        """
        Return a *new* variable.

        INPUT:

        - ``decision`` -- accepted for compatibility with other solvers; ignored

        EXAMPLES::

            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: solver = DIMACS()
            sage: solver.var()
            1
        """
        self._var += 1
        return self._var

    def nvars(self):
        """
        Return the number of variables.

        EXAMPLES::

            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: solver = DIMACS()
            sage: solver.var()
            1
            sage: solver.var(decision=True)
            2
            sage: solver.nvars()
            2
        """
        return self._var

    def add_clause(self, lits):
        """
        Add a new clause to set of clauses.

        INPUT:

        - ``lits`` -- tuple of nonzero integers

        .. NOTE::

            If any element ``e`` in ``lits`` has ``abs(e)`` greater
            than the number of variables generated so far, then new
            variables are created automatically.

        EXAMPLES::

            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: solver = DIMACS()
            sage: solver.var()
            1
            sage: solver.var(decision=True)
            2
            sage: solver.add_clause( (1, -2 , 3) )
            sage: solver
            DIMACS Solver: ''
        """
        l = []
        for lit in lits:
            lit = int(lit)
            while abs(lit) > self.nvars():
                self.var()
            l.append(str(lit))
        l.append("0\n")
        self._tail.write(" ".join(l))
        self._lit += 1

    def write(self, filename=None):
        """
        Write DIMACS file.

        INPUT:

        - ``filename`` -- if ``None`` default filename specified at initialization is used for
          writing to (default: ``None``)

        EXAMPLES::

            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: fn = tmp_filename()
            sage: solver = DIMACS(filename=fn)
            sage: solver.add_clause( (1, -2 , 3) )
            sage: _ = solver.write()
            sage: for line in open(fn).readlines():
            ....:     print(line)
            p cnf 3 1
            1 -2 3 0

            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: fn = tmp_filename()
            sage: solver = DIMACS()
            sage: solver.add_clause( (1, -2 , 3) )
            sage: _ = solver.write(fn)
            sage: for line in open(fn).readlines():
            ....:      print(line)
            p cnf 3 1
            1 -2 3 0
        """
        headname = self._headname if filename is None else filename
        head = open(headname, "w")
        head.truncate(0)
        head.write("p cnf %d %d\n" % (self._var, self._lit))
        head.close()

        tail = self._tail
        tail.close()

        head = open(headname, "a")
        tail = open(self._tail.name)
        head.write(tail.read())
        tail.close()
        head.close()

        self._tail = open(self._tail.name, "a")
        return headname

    def clauses(self, filename=None):
        """
        Return original clauses.

        INPUT:

        - ``filename`` -- if not ``None`` clauses are written to ``filename`` in
          DIMACS format (default: ``None``)

        OUTPUT:

            If ``filename`` is ``None`` then a list of ``lits, is_xor, rhs``
            tuples is returned, where ``lits`` is a tuple of literals,
            ``is_xor`` is always ``False`` and ``rhs`` is always ``None``.

            If ``filename`` points to a writable file, then the list of original
            clauses is written to that file in DIMACS format.

        EXAMPLES::

            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: fn = tmp_filename()
            sage: solver = DIMACS()
            sage: solver.add_clause( (1, 2, 3) )
            sage: solver.clauses()
            [((1, 2, 3), False, None)]

            sage: solver.add_clause( (1, 2, -3) )
            sage: solver.clauses(fn)
            sage: print(open(fn).read())
            p cnf 3 2
            1 2 3 0
            1 2 -3 0
            <BLANKLINE>
        """
        if filename is not None:
            self.write(filename)
        else:
            tail = self._tail
            tail.close()
            tail = open(self._tail.name)

            clauses = []
            for line in tail.readlines():
                if line.startswith("p") or line.startswith("c"):
                    continue
                clause = []
                for lit in line.split(" "):
                    lit = int(lit)
                    if lit == 0:
                        break
                    clause.append(lit)
                clauses.append((tuple(clause), False, None))
            tail.close()
            self._tail = open(self._tail.name, "a")
            return clauses

    @staticmethod
    def render_dimacs(clauses, filename, nlits):
        """
        Produce DIMACS file ``filename`` from ``clauses``.

        INPUT:

        - ``clauses`` -- list of clauses, either in simple format as a list of
          literals or in extended format for CryptoMiniSat: a tuple of literals,
          ``is_xor`` and ``rhs``.

        - ``filename`` -- the file to write to

        - ``nlits -- the number of literals appearing in ``clauses``

        EXAMPLES::

            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: fn = tmp_filename()
            sage: solver = DIMACS()
            sage: solver.add_clause( (1, 2, -3) )
            sage: DIMACS.render_dimacs(solver.clauses(), fn, solver.nvars())
            sage: print(open(fn).read())
            p cnf 3 1
            1 2 -3 0
            <BLANKLINE>

        This is equivalent to::

            sage: solver.clauses(fn)
            sage: print(open(fn).read())
            p cnf 3 1
            1 2 -3 0
            <BLANKLINE>

        This function also accepts a "simple" format::

            sage: DIMACS.render_dimacs([ (1,2), (1,2,-3) ], fn, 3)
            sage: print(open(fn).read())
            p cnf 3 2
            1 2 0
            1 2 -3 0
            <BLANKLINE>
        """
        with open(filename, "w") as fh:
            fh.write("p cnf %d %d\n" % (nlits, len(clauses)))
            for clause in clauses:
                if len(clause) == 3 and clause[1] in (True, False) and clause[2] in (True, False, None):
                    lits, is_xor, rhs = clause
                else:
                    lits, is_xor, rhs = clause, False, None

                if is_xor:
                    closing = lits[-1] if rhs else -lits[-1]
                    fh.write("x" + " ".join(map(str, lits[:-1])) + " %d 0\n" % closing)
                else:
                    fh.write(" ".join(map(str, lits)) + " 0\n")

    def _run(self):
        r"""
        Run 'command' and collect output.

        TESTS:

        This class is not meant to be called directly::

            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: fn = tmp_filename()
            sage: solver = DIMACS(filename=fn)
            sage: solver.add_clause( (1, -2 , 3) )
            sage: solver._run()
            Traceback (most recent call last):
            ...
            ValueError: no SAT solver command selected

        It is used by subclasses::

            sage: from sage.sat.solvers import Glucose
            sage: solver = Glucose()
            sage: solver.add_clause( (1, 2, 3) )
            sage: solver.add_clause( (-1,) )
            sage: solver.add_clause( (-2,) )
            sage: solver._run()                       # optional - glucose
            sage: solver._output                      # optional - glucose
            [...
             's SATISFIABLE\n',
             'v -1 -2 3 0\n']
        """
        from sage.misc.verbose import get_verbose

        self.write()
        output_filename = None
        self._output = []

        command = self._command

        if not command:
            raise ValueError("no SAT solver command selected")

        if "{output}" in command:
            output_filename = tmp_filename()

        if isinstance(command, str):
            command = command.strip().format(input=self._headname, output=output_filename)
            args = shlex.split(command)
        else:
            args = [(arg.format(input=self._headname, output=output_filename)
                     if isinstance(arg, str)
                     else arg.absolute_filename())
                    for arg in command]

        try:
            process = subprocess.Popen(args, stdout=subprocess.PIPE)
        except OSError:
            raise OSError("Could not run '%s', perhaps you need to add your SAT solver to $PATH?" % (" ".join(args)))

        try:
            while process.poll() is None:
                for line in iter(process.stdout.readline, b''):
                    if get_verbose() or self._verbosity:
                        print(line)
                        sys.stdout.flush()
                    self._output.append(line.decode('utf-8'))
                sleep(0.1)
            if output_filename:
                self._output.extend(open(output_filename).readlines())
        except BaseException:
            process.kill()
            raise

    def __call__(self, assumptions=None):
        """
        Solve this instance and return the parsed output.

        INPUT:

        - ``assumptions`` -- ignored, accepted for compatibility with
          other solvers (default: ``None``)

        OUTPUT:

        - If this instance is SAT: A tuple of length ``nvars()+1``
          where the ``i``-th entry holds an assignment for the
          ``i``-th variables (the ``0``-th entry is always ``None``).

        - If this instance is UNSAT: ``False``

        EXAMPLES:

        When the problem is SAT::

            sage: from sage.sat.solvers import RSat
            sage: solver = RSat()
            sage: solver.add_clause( (1, 2, 3) )
            sage: solver.add_clause( (-1,) )
            sage: solver.add_clause( (-2,) )
            sage: solver()                            # optional - rsat
            (None, False, False, True)

        When the problem is UNSAT::

            sage: solver = RSat()
            sage: solver.add_clause((1,2))
            sage: solver.add_clause((-1,2))
            sage: solver.add_clause((1,-2))
            sage: solver.add_clause((-1,-2))
            sage: solver()                            # optional - rsat
            False

        With Glucose::

            sage: from sage.sat.solvers.dimacs import Glucose
            sage: solver = Glucose()
            sage: solver.add_clause((1,2))
            sage: solver.add_clause((-1,2))
            sage: solver.add_clause((1,-2))
            sage: solver()                           # optional - glucose
            (None, True, True)
            sage: solver.add_clause((-1,-2))
            sage: solver()                           # optional - glucose
            False

        With GlucoseSyrup::

            sage: from sage.sat.solvers.dimacs import GlucoseSyrup
            sage: solver = GlucoseSyrup()
            sage: solver.add_clause((1,2))
            sage: solver.add_clause((-1,2))
            sage: solver.add_clause((1,-2))
            sage: solver()                          # optional - glucose
            (None, True, True)
            sage: solver.add_clause((-1,-2))
            sage: solver()                          # optional - glucose
            False

        TESTS::

            sage: from sage.sat.boolean_polynomials import solve as solve_sat
            sage: sr = mq.SR(1, 1, 1, 4, gf2=True, polybori=True)                       # needs sage.rings.finite_rings sage.rings.polynomial.pbori
            sage: while True:  # workaround (see :issue:`31891`)                         # needs sage.rings.finite_rings sage.rings.polynomial.pbori
            ....:     try:
            ....:         F, s = sr.polynomial_system()
            ....:         break
            ....:     except ZeroDivisionError:
            ....:         pass
            sage: solve_sat(F, solver=sage.sat.solvers.RSat)    # optional - rsat, needs sage.rings.finite_rings sage.rings.polynomial.pbori
        """
        if assumptions is not None:
            raise NotImplementedError("Assumptions are not supported for DIMACS based solvers.")

        self._run()

        v_lines = []
        for line in self._output:
            if line.startswith("c"):
                continue
            if line.startswith("s"):
                if "UNSAT" in line:
                    return False
            if line.startswith("v"):
                v_lines.append(line[1:].strip())

        if v_lines:
            L = " ".join(v_lines).split(" ")
            assert L[-1] == "0", "last digit of solution line must be zero (not {})".format(L[-1])
            return (None,) + tuple(int(e) > 0 for e in L[:-1])
        else:
            raise ValueError("When parsing the output(={}), no line starts with letter v or s".format(self._output))


class RSat(DIMACS):
    """
    An instance of the RSat solver.

    For information on RSat see: http://reasoning.cs.ucla.edu/rsat/

    EXAMPLES::

        sage: from sage.sat.solvers import RSat
        sage: solver = RSat()
        sage: solver
        DIMACS Solver: 'rsat {input} -v -s'

    When the problem is SAT::

        sage: from sage.sat.solvers import RSat
        sage: solver = RSat()
        sage: solver.add_clause( (1, 2, 3) )
        sage: solver.add_clause( (-1,) )
        sage: solver.add_clause( (-2,) )
        sage: solver()                            # optional - rsat
        (None, False, False, True)

    When the problem is UNSAT::

        sage: solver = RSat()
        sage: solver.add_clause((1,2))
        sage: solver.add_clause((-1,2))
        sage: solver.add_clause((1,-2))
        sage: solver.add_clause((-1,-2))
        sage: solver()                            # optional - rsat
        False
    """
    command = ["rsat", "{input}", "-v", "-s"]


class Glucose(DIMACS):
    """
    An instance of the Glucose solver.

    For information on Glucose see: http://www.labri.fr/perso/lsimon/glucose/

    EXAMPLES::

        sage: from sage.sat.solvers import Glucose
        sage: solver = Glucose()
        sage: solver
        DIMACS Solver: 'glucose -verb=0 -model {input}'

    When the problem is SAT::

        sage: from sage.sat.solvers import Glucose
        sage: solver1 = Glucose()
        sage: solver1.add_clause( (1, 2, 3) )
        sage: solver1.add_clause( (-1,) )
        sage: solver1.add_clause( (-2,) )
        sage: solver1()                            # optional - glucose
        (None, False, False, True)

    When the problem is UNSAT::

        sage: solver2 = Glucose()
        sage: solver2.add_clause((1,2))
        sage: solver2.add_clause((-1,2))
        sage: solver2.add_clause((1,-2))
        sage: solver2.add_clause((-1,-2))
        sage: solver2()                            # optional - glucose
        False

    With one hundred variables::

        sage: solver3 = Glucose()
        sage: solver3.add_clause( (1, 2, 100) )
        sage: solver3.add_clause( (-1,) )
        sage: solver3.add_clause( (-2,) )
        sage: solver3()                            # optional - glucose
        (None, False, False, ..., True)

    TESTS::

        sage: print(''.join(solver1._output))      # optional - glucose
        c...
        s SATISFIABLE
        v -1 -2 3 0

    ::

        sage: print(''.join(solver2._output))      # optional - glucose
        c...
        s UNSATISFIABLE

    Glucose gives large solution on one single line::

        sage: print(''.join(solver3._output))      # optional - glucose
        c...
        s SATISFIABLE
        v -1 -2 ... 100 0
    """
    command = [Glucose_executable(), "-verb=0", "-model", "{input}"]


class GlucoseSyrup(DIMACS):
    """
    An instance of the Glucose-syrup parallel solver.

    For information on Glucose see: http://www.labri.fr/perso/lsimon/glucose/

    EXAMPLES::

        sage: from sage.sat.solvers import GlucoseSyrup
        sage: solver = GlucoseSyrup()
        sage: solver
        DIMACS Solver: 'glucose-syrup -model -verb=0 {input}'

    When the problem is SAT::

        sage: solver1 = GlucoseSyrup()
        sage: solver1.add_clause( (1, 2, 3) )
        sage: solver1.add_clause( (-1,) )
        sage: solver1.add_clause( (-2,) )
        sage: solver1()                            # optional - glucose
        (None, False, False, True)

    When the problem is UNSAT::

        sage: solver2 = GlucoseSyrup()
        sage: solver2.add_clause((1,2))
        sage: solver2.add_clause((-1,2))
        sage: solver2.add_clause((1,-2))
        sage: solver2.add_clause((-1,-2))
        sage: solver2()                            # optional - glucose
        False

    With one hundred variables::

        sage: solver3 = GlucoseSyrup()
        sage: solver3.add_clause( (1, 2, 100) )
        sage: solver3.add_clause( (-1,) )
        sage: solver3.add_clause( (-2,) )
        sage: solver3()                            # optional - glucose
        (None, False, False, ..., True)

    TESTS::

        sage: print(''.join(solver1._output))      # optional - glucose
        c...
        s SATISFIABLE
        v -1 -2 3 0

    ::

        sage: print(''.join(solver2._output))      # optional - glucose
        c...
        s UNSATISFIABLE

    GlucoseSyrup gives large solution on one single line::

        sage: print(''.join(solver3._output))      # optional - glucose
        c...
        s SATISFIABLE
        v -1 -2 ... 100 0
    """
    command = [Glucose_executable("glucose-syrup"), "-model", "-verb=0", "{input}"]


class Kissat(DIMACS):
    """
    An instance of the Kissat SAT solver.

    For information on Kissat see: http://fmv.jku.at/kissat/

    EXAMPLES::

        sage: from sage.sat.solvers import Kissat
        sage: solver = Kissat()
        sage: solver
        DIMACS Solver: 'kissat -q {input}'

    When the problem is SAT::

        sage: solver1 = Kissat()
        sage: solver1.add_clause( (1, 2, 3) )
        sage: solver1.add_clause( (-1,) )
        sage: solver1.add_clause( (-2,) )
        sage: solver1()                           # optional - kissat
        (None, False, False, True)

    When the problem is UNSAT::

        sage: solver2 = Kissat()
        sage: solver2.add_clause((1,2))
        sage: solver2.add_clause((-1,2))
        sage: solver2.add_clause((1,-2))
        sage: solver2.add_clause((-1,-2))
        sage: solver2()                           # optional - kissat
        False

    With one hundred variables::

        sage: solver3 = Kissat()
        sage: solver3.add_clause( (1, 2, 100) )
        sage: solver3.add_clause( (-1,) )
        sage: solver3.add_clause( (-2,) )
        sage: solver3()                           # optional - kissat
        (None, False, False, ..., True)

    TESTS::

        sage: print(''.join(solver1._output))     # optional - kissat
        s SATISFIABLE
        v -1 -2 3 0

    ::

        sage: print(''.join(solver2._output))     # optional - kissat
        s UNSATISFIABLE

    Here the output contains many lines starting with letter "v"::

        sage: print(''.join(solver3._output))     # optional - kissat
        s SATISFIABLE
        v -1 -2 ...
        v ...
        v ...
        v ... 100 0
    """

    command = [Kissat_executable(), "-q", "{input}"]
